# TODO: this is rendering mostly ported or adapted from the original Minigrid. A bit dirty right now...
import functools
import math
from typing import Optional, Union

import numpy as np

from ..core.constants import Colors, Tiles
from ..types import AgentState, IntOrArray
from .utils import (
    downsample,
    fill_coords,
    highlight_img,
    point_in_circle,
    point_in_hexagon,
    point_in_rect,
    point_in_triangle,
    rotate_fn,
)

COLORS_MAP = {
    Colors.EMPTY: np.array((255, 255, 255)),  # just a placeholder
    Colors.RED: np.array((255, 0, 0)),
    Colors.GREEN: np.array((0, 255, 0)),
    Colors.BLUE: np.array((0, 0, 255)),
    Colors.PURPLE: np.array((112, 39, 195)),
    Colors.YELLOW: np.array((255, 255, 0)),
    Colors.GREY: np.array((100, 100, 100)),
    Colors.BLACK: np.array((0, 0, 0)),
    Colors.ORANGE: np.array((255, 140, 0)),
    Colors.WHITE: np.array((255, 255, 255)),
    Colors.BROWN: np.array((160, 82, 45)),
    Colors.PINK: np.array((225, 20, 147)),
}


def _render_empty(img: np.ndarray, color: int):
    fill_coords(img, point_in_rect(0.45, 0.55, 0.2, 0.65), COLORS_MAP[Colors.RED])
    fill_coords(img, point_in_rect(0.45, 0.55, 0.7, 0.85), COLORS_MAP[Colors.RED])

    fill_coords(img, point_in_rect(0, 0.031, 0, 1), COLORS_MAP[Colors.RED])
    fill_coords(img, point_in_rect(0, 1, 0, 0.031), COLORS_MAP[Colors.RED])
    fill_coords(img, point_in_rect(1 - 0.031, 1, 0, 1), COLORS_MAP[Colors.RED])
    fill_coords(img, point_in_rect(0, 1, 1 - 0.031, 1), COLORS_MAP[Colors.RED])


def _render_floor(img: np.ndarray, color: int):
    # draw the grid lines (top and left edges)
    fill_coords(img, point_in_rect(0, 0.031, 0, 1), (100, 100, 100))
    fill_coords(img, point_in_rect(0, 1, 0, 0.031), (100, 100, 100))
    # draw tile
    fill_coords(img, point_in_rect(0.031, 1, 0.031, 1), COLORS_MAP[color] / 2)

    # # other grid lines (was used for paper visualizations)
    # fill_coords(img, point_in_rect(1 - 0.031, 1, 0, 1), (100, 100, 100))
    # fill_coords(img, point_in_rect(0, 1, 1 - 0.031, 1), (100, 100, 100))
    #


def _render_wall(img: np.ndarray, color: int):
    fill_coords(img, point_in_rect(0, 1, 0, 1), COLORS_MAP[color])


def _render_ball(img: np.ndarray, color: int):
    _render_floor(img, Colors.BLACK)
    fill_coords(img, point_in_circle(0.5, 0.5, 0.31), COLORS_MAP[color])


def _render_square(img: np.ndarray, color: int):
    _render_floor(img, Colors.BLACK)
    fill_coords(img, point_in_rect(0.25, 0.75, 0.25, 0.75), COLORS_MAP[color])


def _render_pyramid(img: np.ndarray, color: int):
    _render_floor(img, Colors.BLACK)
    tri_fn = point_in_triangle(
        (0.15, 0.8),
        (0.85, 0.8),
        (0.5, 0.2),
    )
    fill_coords(img, tri_fn, COLORS_MAP[color])


def _render_hex(img: np.ndarray, color: int):
    _render_floor(img, Colors.BLACK)
    fill_coords(img, point_in_hexagon(0.35), COLORS_MAP[color])


def _render_star(img: np.ndarray, color: int):
    # yes, this is a hexagram not a star, but who cares...
    _render_floor(img, Colors.BLACK)
    tri_fn2 = point_in_triangle(
        (0.15, 0.75),
        (0.85, 0.75),
        (0.5, 0.15),
    )
    tri_fn1 = point_in_triangle(
        (0.15, 0.3),
        (0.85, 0.3),
        (0.5, 0.9),
    )
    fill_coords(img, tri_fn1, COLORS_MAP[color])
    fill_coords(img, tri_fn2, COLORS_MAP[color])


def _render_goal(img: np.ndarray, color: int):
    # draw the grid lines (top and left edges)
    fill_coords(img, point_in_rect(0, 0.031, 0, 1), (100, 100, 100))
    fill_coords(img, point_in_rect(0, 1, 0, 0.031), (100, 100, 100))
    # draw tile
    fill_coords(img, point_in_rect(0.031, 1, 0.031, 1), COLORS_MAP[color])

    # # other grid lines (was used for paper visualizations)
    # fill_coords(img, point_in_rect(1 - 0.031, 1, 0, 1), (100, 100, 100))
    # fill_coords(img, point_in_rect(0, 1, 1 - 0.031, 1), (100, 100, 100))


def _render_key(img: np.ndarray, color: int):
    _render_floor(img, Colors.BLACK)
    # Vertical quad
    fill_coords(img, point_in_rect(0.50, 0.63, 0.31, 0.88), COLORS_MAP[color])
    # Teeth
    fill_coords(img, point_in_rect(0.38, 0.50, 0.59, 0.66), COLORS_MAP[color])
    fill_coords(img, point_in_rect(0.38, 0.50, 0.81, 0.88), COLORS_MAP[color])
    # Ring
    fill_coords(img, point_in_circle(cx=0.56, cy=0.28, r=0.190), COLORS_MAP[color])
    fill_coords(img, point_in_circle(cx=0.56, cy=0.28, r=0.064), (0, 0, 0))


def _render_door_locked(img: np.ndarray, color: int):
    fill_coords(img, point_in_rect(0.00, 1.00, 0.00, 1.00), COLORS_MAP[color])
    fill_coords(img, point_in_rect(0.06, 0.94, 0.06, 0.94), 0.45 * COLORS_MAP[color])
    # Draw key slot
    fill_coords(img, point_in_rect(0.52, 0.75, 0.50, 0.56), COLORS_MAP[color])


def _render_door_closed(img: np.ndarray, color: int):
    fill_coords(img, point_in_rect(0.00, 1.00, 0.00, 1.00), COLORS_MAP[color])
    fill_coords(img, point_in_rect(0.04, 0.96, 0.04, 0.96), (0, 0, 0))
    fill_coords(img, point_in_rect(0.08, 0.92, 0.08, 0.92), COLORS_MAP[color])
    fill_coords(img, point_in_rect(0.12, 0.88, 0.12, 0.88), (0, 0, 0))
    # Draw door handle
    fill_coords(img, point_in_circle(cx=0.75, cy=0.50, r=0.08), COLORS_MAP[color])


def _render_door_open(img: np.ndarray, color: int):
    _render_floor(img, Colors.BLACK)
    # draw the grid lines (top and left edges)
    fill_coords(img, point_in_rect(0, 0.031, 0, 1), (100, 100, 100))
    fill_coords(img, point_in_rect(0, 1, 0, 0.031), (100, 100, 100))
    # draw door
    fill_coords(img, point_in_rect(0.88, 1.00, 0.00, 1.00), COLORS_MAP[color])
    fill_coords(img, point_in_rect(0.92, 0.96, 0.04, 0.96), (0, 0, 0))


def _render_player(img: np.ndarray, direction: int):
    tri_fn = point_in_triangle(
        (0.12, 0.19),
        (0.87, 0.50),
        (0.12, 0.81),
    )
    # Rotate the agent based on its direction
    tri_fn = rotate_fn(tri_fn, cx=0.5, cy=0.5, theta=0.5 * math.pi * (direction - 1))
    fill_coords(img, tri_fn, COLORS_MAP[Colors.RED])


TILES_FN_MAP = {
    Tiles.FLOOR: _render_floor,
    Tiles.WALL: _render_wall,
    Tiles.BALL: _render_ball,
    Tiles.SQUARE: _render_square,
    Tiles.PYRAMID: _render_pyramid,
    Tiles.HEX: _render_hex,
    Tiles.STAR: _render_star,
    Tiles.GOAL: _render_goal,
    Tiles.KEY: _render_key,
    Tiles.DOOR_LOCKED: _render_door_locked,
    Tiles.DOOR_CLOSED: _render_door_closed,
    Tiles.DOOR_OPEN: _render_door_open,
    Tiles.EMPTY: _render_empty,
}


# TODO: add highlight for can_see_through_walls=Fasle
def get_highlight_mask(grid: np.ndarray, agent: Optional[AgentState], view_size: int) -> np.ndarray:
    mask = np.zeros((grid.shape[0] + 2 * view_size, grid.shape[1] + 2 * view_size), dtype=np.bool_)
    if agent is None:
        return mask

    agent_y, agent_x = agent.position + view_size
    if agent.direction == 0:
        y, x = agent_y - view_size + 1, agent_x - (view_size // 2)
    elif agent.direction == 1:
        y, x = agent_y - (view_size // 2), agent_x
    elif agent.direction == 2:
        y, x = agent_y, agent_x - (view_size // 2)
    elif agent.direction == 3:
        y, x = agent_y - (view_size // 2), agent_x - view_size + 1
    else:
        raise RuntimeError("Unknown direction")

    mask[y : y + view_size, x : x + view_size] = True
    mask = mask[view_size:-view_size, view_size:-view_size]
    assert mask.shape == (grid.shape[0], grid.shape[1])

    return mask


@functools.cache
def render_tile(
    tile: tuple, agent_direction: Optional[int] = None, highlight: bool = False, tile_size: int = 32, subdivs: int = 3
) -> np.ndarray:
    img = np.full((tile_size * subdivs, tile_size * subdivs, 3), dtype=np.uint8, fill_value=255)
    # draw tile
    TILES_FN_MAP[tile[0]](img, tile[1])
    # draw agent if on this tile
    if agent_direction is not None:
        _render_player(img, agent_direction)

    if highlight:
        highlight_img(img, alpha=0.2)

    # downsample the image to perform supersampling/anti-aliasing
    img = downsample(img, subdivs)

    return img


# WARN: will NOT work under jit and needed for debugging/presentation mainly.
def render(
    grid: np.ndarray,
    agent: Optional[AgentState] = None,
    view_size: IntOrArray = 7,
    tile_size: IntOrArray = 32,
) -> np.ndarray:
    # grid = np.asarray(grid)
    # compute the total grid size
    height_px = grid.shape[0] * int(tile_size)
    width_px = grid.shape[1] * int(tile_size)

    img = np.full((height_px, width_px, 3), dtype=np.uint8, fill_value=255)

    # compute agent fov highlighting
    highlight_mask = get_highlight_mask(grid, agent, int(view_size))

    for y in range(grid.shape[0]):
        for x in range(grid.shape[1]):
            if agent is not None and np.array_equal((y, x), agent.position):
                agent_direction = int(agent.direction)
            else:
                agent_direction = None

            tile_img = render_tile(
                tile=tuple(grid[y, x].tolist()),
                agent_direction=agent_direction,
                highlight=highlight_mask[y, x],
                tile_size=int(tile_size),
            )

            ymin = y * int(tile_size)
            ymax = (y + 1) * int(tile_size)
            xmin = x * int(tile_size)
            xmax = (x + 1) * int(tile_size)
            img[ymin:ymax, xmin:xmax, :] = tile_img

    return img



from __future__ import annotations

from typing import Callable, Optional, Union

import jax
import jax.numpy as jnp

from ..types import GridState, IntOrArray, Tile
from .constants import FREE_TO_PUT_DOWN, LOS_BLOCKING, PICKABLE, TILES_REGISTRY, WALKABLE, Colors, Tiles


def empty_world(height: IntOrArray, width: IntOrArray) -> GridState:
    grid = jnp.zeros((height, width, 2), dtype=jnp.uint8)
    grid = grid.at[:, :, 0:2].set(TILES_REGISTRY[Tiles.FLOOR, Colors.BLACK])
    return grid


def equal(tile1: Tile, tile2: Tile) -> Tile:
    # wait, is this just a jnp.array_equal?
    return jnp.all(jnp.equal(tile1, tile2))


def get_neighbouring_tiles(grid: GridState, y: IntOrArray, x: IntOrArray) -> tuple[Tile, Tile, Tile, Tile]:
    # end_of_map = TILES_REGISTRY[Tiles.END_OF_MAP, Colors.END_OF_MAP]
    end_of_map = Tiles.EMPTY

    up_tile = grid.at[y - 1, x].get(mode="fill", fill_value=end_of_map)
    right_tile = grid.at[y, x + 1].get(mode="fill", fill_value=end_of_map)
    down_tile = grid.at[y + 1, x].get(mode="fill", fill_value=end_of_map)
    left_tile = grid.at[y, x - 1].get(mode="fill", fill_value=end_of_map)
    # up_tile = jax.lax.select(y > 0, grid[y - 1, x], end_of_map)
    # right_tile = jax.lax.select(x < grid.shape[1] - 2, grid[y, x + 1], end_of_map)
    # down_tile = jax.lax.select(y < grid.shape[0] - 2, grid[y + 1, x], end_of_map)
    # left_tile = jax.lax.select(x > 0, grid[y, x - 1], end_of_map)
    return up_tile, right_tile, down_tile, left_tile


def horizontal_line(grid: GridState, x: IntOrArray, y: IntOrArray, length: IntOrArray, tile: Tile) -> GridState:
    grid = grid.at[y, x : x + length].set(tile)
    return grid


def vertical_line(grid: GridState, x: IntOrArray, y: IntOrArray, length: IntOrArray, tile: Tile) -> GridState:
    grid = grid.at[y : y + length, x].set(tile)
    return grid


def rectangle(grid: GridState, x: IntOrArray, y: IntOrArray, h: IntOrArray, w: IntOrArray, tile: Tile) -> GridState:
    grid = vertical_line(grid, x, y, h, tile)
    grid = vertical_line(grid, x + w - 1, y, h, tile)
    grid = horizontal_line(grid, x, y, w, tile)
    grid = horizontal_line(grid, x, y + h - 1, w, tile)
    return grid


def room(height: IntOrArray, width: IntOrArray) -> GridState:
    grid = empty_world(height, width)
    grid = rectangle(grid, 0, 0, height, width, tile=TILES_REGISTRY[Tiles.WALL, Colors.GREY])
    return grid


def two_rooms(height: IntOrArray, width: IntOrArray) -> GridState:
    wall_tile: Tile = TILES_REGISTRY[Tiles.WALL, Colors.GREY]

    grid = empty_world(height, width)
    grid = rectangle(grid, 0, 0, height, width, tile=wall_tile)
    grid = vertical_line(grid, width // 2, 0, height, tile=wall_tile)
    return grid


def four_rooms(height: IntOrArray, width: IntOrArray) -> GridState:
    wall_tile: Tile = TILES_REGISTRY[Tiles.WALL, Colors.GREY]

    grid = empty_world(height, width)
    grid = rectangle(grid, 0, 0, height, width, tile=wall_tile)
    grid = vertical_line(grid, width // 2, 0, height, tile=wall_tile)
    grid = horizontal_line(grid, 0, height // 2, width, tile=wall_tile)
    return grid


def nine_rooms(height: IntOrArray, width: IntOrArray) -> GridState:
    wall_tile: Tile = TILES_REGISTRY[Tiles.WALL, Colors.GREY]

    grid = empty_world(height, width)
    grid = rectangle(grid, 0, 0, height, width, tile=wall_tile)
    grid = vertical_line(grid, width // 3, 0, height, tile=wall_tile)
    grid = vertical_line(grid, 2 * (width // 3), 0, height, tile=wall_tile)
    grid = horizontal_line(grid, 0, height // 3, width, tile=wall_tile)
    grid = horizontal_line(grid, 0, 2 * (height // 3), width, tile=wall_tile)
    return grid


def check_walkable(grid: GridState, position: jax.Array) -> jax.Array:
    tile_id = grid[position[0], position[1], 0]
    is_walkable = jnp.isin(tile_id, WALKABLE, assume_unique=True)

    return is_walkable


def check_pickable(grid: GridState, position: jax.Array) -> jax.Array:
    tile_id = grid[position[0], position[1], 0]
    is_pickable = jnp.isin(tile_id, PICKABLE, assume_unique=True)
    return is_pickable


def check_can_put(grid: GridState, position: jax.Array) -> jax.Array:
    tile_id = grid[position[0], position[1], 0]
    can_put = jnp.isin(tile_id, FREE_TO_PUT_DOWN, assume_unique=True)

    return can_put


def check_see_behind(grid: GridState, position: jax.Array) -> jax.Array:
    tile_id = grid[position[0], position[1], 0]
    is_not_blocking = jnp.isin(tile_id, LOS_BLOCKING, assume_unique=True, invert=True)

    return is_not_blocking


def align_with_up(grid: GridState, direction: IntOrArray) -> GridState:
    aligned_grid = jax.lax.switch(
        direction,
        (
            lambda: grid,
            lambda: jnp.rot90(grid, 1),
            lambda: jnp.rot90(grid, 2),
            lambda: jnp.rot90(grid, 3),
        ),
    )
    return aligned_grid


def grid_coords(grid: GridState) -> jax.Array:
    coords = jnp.mgrid[: grid.shape[0], : grid.shape[1]]
    coords = coords.transpose(1, 2, 0).reshape(-1, 2)
    return coords


def transparent_mask(grid: GridState) -> jax.Array:
    coords = grid_coords(grid)
    mask = jax.vmap(check_see_behind, in_axes=(None, 0))(grid, coords)
    mask = mask.reshape(grid.shape[0], grid.shape[1])
    return mask


def free_tiles_mask(grid: GridState) -> jax.Array:
    coords = grid_coords(grid)
    mask = jax.vmap(check_can_put, in_axes=(None, 0))(grid, coords)
    mask = mask.reshape(grid.shape[0], grid.shape[1])
    return mask


def coordinates_mask(grid: GridState, address: tuple[IntOrArray, IntOrArray], comparison_fn: Callable) -> jax.Array:
    positions = jnp.mgrid[: grid.shape[0], : grid.shape[1]]
    cond_1 = comparison_fn(positions[0], address[0])
    cond_2 = comparison_fn(positions[1], address[1])
    mask = jnp.logical_and(cond_1, cond_2)
    return mask


def sample_coordinates(key: jax.Array, grid: GridState, num: int, mask: Optional[jax.Array] = None) -> jax.Array:
    if mask is None:
        mask = jnp.ones((grid.shape[0], grid.shape[1]), dtype=jnp.bool_)

    coords = jax.random.choice(
        key=key,
        shape=(num,),
        a=jnp.arange(grid.shape[0] * grid.shape[1]),
        replace=False,
        p=(mask & free_tiles_mask(grid)).flatten(),
    )
    coords = jnp.divmod(coords, grid.shape[1])
    coords = jnp.concatenate((coords[0].reshape(-1, 1), coords[1].reshape(-1, 1)), axis=-1)
    return coords


def sample_direction(key: jax.Array) -> jax.Array:
    return jax.random.randint(key, shape=(), minval=0, maxval=4)


def pad_along_axis(arr: jax.Array, pad_to: int, axis: int = 0, fill_value: int = 0) -> jax.Array:
    pad_size = pad_to - arr.shape[axis]
    if pad_size <= 0:
        return arr

    # manually annotate for pyright
    npad: list[tuple[int, int]] = [(0, 0)] * arr.ndim
    npad[axis] = (0, pad_size)
    return jnp.pad(arr, pad_width=npad, mode="constant", constant_values=fill_value)


def cartesian_product_1d(a: jax.Array, b: jax.Array) -> jax.Array:
    return jnp.dstack(jnp.meshgrid(a, b)).reshape(-1, 2)



from .benchmarks import load_benchmark, registered_benchmarks
from .registration import make, register, registered_environments

# TODO: add __all__
__version__ = "0.9.3"

# ---------- XLand-MiniGrid environments ----------

# WARN: TMP, only for FPS measurements, will remove later
# register(
#     id="MiniGrid-1Rules",
#     entry_point="xminigrid.envs.xland_tmp:XLandMiniGrid",
#     num_rules=1,
#     height=16,
#     width=16,
# )
#
# register(
#     id="MiniGrid-3Rules",
#     entry_point="xminigrid.envs.xland_tmp:XLandMiniGrid",
#     num_rules=2,
#     height=16,
#     width=16,
# )
#
# register(
#     id="MiniGrid-6Rules",
#     entry_point="xminigrid.envs.xland_tmp:XLandMiniGrid",
#     num_rules=6,
#     height=16,
#     width=16,
# )
#
# register(
#     id="MiniGrid-12Rules",
#     entry_point="xminigrid.envs.xland_tmp:XLandMiniGrid",
#     num_rules=12,
#     height=16,
#     width=16,
# )
#
# register(
#     id="MiniGrid-24Rules",
#     entry_point="xminigrid.envs.xland_tmp:XLandMiniGrid",
#     num_rules=24,
#     height=16,
#     width=16,
# )

# register(
#     id="XLand-MiniGrid-R1-8x8",
#     entry_point="xminigrid.envs.xland:XLandMiniGrid",
#     grid_type="R1",
#     height=8,
#     width=8,
# )
#
# register(
#     id="XLand-MiniGrid-R1-16x16",
#     entry_point="xminigrid.envs.xland:XLandMiniGrid",
#     grid_type="R1",
#     height=16,
#     width=16,
# )
#
# register(
#     id="XLand-MiniGrid-R1-32x32",
#     entry_point="xminigrid.envs.xland:XLandMiniGrid",
#     grid_type="R1",
#     height=32,
#     width=32,
# )
#
# register(
#     id="XLand-MiniGrid-R1-64x64",
#     entry_point="xminigrid.envs.xland:XLandMiniGrid",
#     grid_type="R1",
#     height=64,
#     width=64,
# )

register(
    id="XLand-MiniGrid-R1-9x9",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R1",
    height=9,
    width=9,
)

register(
    id="XLand-MiniGrid-R1-11x11",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R1",
    height=11,
    width=11,
)

register(
    id="XLand-MiniGrid-R1-13x13",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R1",
    height=13,
    width=13,
)

register(
    id="XLand-MiniGrid-R1-15x15",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R1",
    height=15,
    width=15,
)

register(
    id="XLand-MiniGrid-R1-17x17",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R1",
    height=17,
    width=17,
)


register(
    id="XLand-MiniGrid-R2-9x9",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R2",
    height=9,
    width=9,
)

register(
    id="XLand-MiniGrid-R2-11x11",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R2",
    height=11,
    width=11,
)

register(
    id="XLand-MiniGrid-R2-13x13",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R2",
    height=13,
    width=13,
)

register(
    id="XLand-MiniGrid-R2-15x15",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R2",
    height=15,
    width=15,
)

register(
    id="XLand-MiniGrid-R2-17x17",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R2",
    height=17,
    width=17,
)


register(
    id="XLand-MiniGrid-R4-9x9",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R4",
    height=9,
    width=9,
)

register(
    id="XLand-MiniGrid-R4-11x11",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R4",
    height=11,
    width=11,
)

register(
    id="XLand-MiniGrid-R4-13x13",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R4",
    height=13,
    width=13,
)

register(
    id="XLand-MiniGrid-R4-15x15",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R4",
    height=15,
    width=15,
)

register(
    id="XLand-MiniGrid-R4-17x17",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R4",
    height=17,
    width=17,
)


register(
    id="XLand-MiniGrid-R6-13x13",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R6",
    height=13,
    width=13,
)

register(
    id="XLand-MiniGrid-R6-17x17",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R6",
    height=17,
    width=17,
)

register(
    id="XLand-MiniGrid-R6-19x19",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R6",
    height=19,
    width=19,
)

# 16, 19, 25
register(
    id="XLand-MiniGrid-R9-16x16",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R9",
    height=16,
    width=16,
)

register(
    id="XLand-MiniGrid-R9-19x19",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R9",
    height=19,
    width=19,
)

register(
    id="XLand-MiniGrid-R9-25x25",
    entry_point="xminigrid.envs.xland:XLandMiniGrid",
    grid_type="R9",
    height=25,
    width=25,
)

# ---------- Environments ported from MiniGrid ----------

# BlockedUnlockPickUp
register(
    id="MiniGrid-BlockedUnlockPickUp",
    entry_point="xminigrid.envs.minigrid.blockedunlockpickup:BlockedUnlockPickUp",
)

# DoorKey
register(
    id="MiniGrid-DoorKey-5x5",
    entry_point="xminigrid.envs.minigrid.doorkey:DoorKey",
    height=5,
    width=5,
)

register(
    id="MiniGrid-DoorKey-6x6",
    entry_point="xminigrid.envs.minigrid.doorkey:DoorKey",
    height=6,
    width=6,
)

register(
    id="MiniGrid-DoorKey-8x8",
    entry_point="xminigrid.envs.minigrid.doorkey:DoorKey",
    height=8,
    width=8,
)

register(
    id="MiniGrid-DoorKey-16x16",
    entry_point="xminigrid.envs.minigrid.doorkey:DoorKey",
    height=16,
    width=16,
)

# Empty
register(
    id="MiniGrid-Empty-5x5",
    entry_point="xminigrid.envs.minigrid.empty:Empty",
    height=5,
    width=5,
)

register(
    id="MiniGrid-Empty-6x6",
    entry_point="xminigrid.envs.minigrid.empty:Empty",
    height=6,
    width=6,
)

register(
    id="MiniGrid-Empty-8x8",
    entry_point="xminigrid.envs.minigrid.empty:Empty",
    height=8,
    width=8,
)

register(
    id="MiniGrid-Empty-16x16",
    entry_point="xminigrid.envs.minigrid.empty:Empty",
    height=16,
    width=16,
)

# EmptyRandom
register(
    id="MiniGrid-EmptyRandom-5x5",
    entry_point="xminigrid.envs.minigrid.empty:EmptyRandom",
    height=5,
    width=5,
)

register(
    id="MiniGrid-EmptyRandom-6x6",
    entry_point="xminigrid.envs.minigrid.empty:EmptyRandom",
    height=6,
    width=6,
)

register(
    id="MiniGrid-EmptyRandom-8x8",
    entry_point="xminigrid.envs.minigrid.empty:EmptyRandom",
    height=8,
    width=8,
)

register(
    id="MiniGrid-EmptyRandom-16x16",
    entry_point="xminigrid.envs.minigrid.empty:EmptyRandom",
    height=16,
    width=16,
)

# FourRooms
register(
    id="MiniGrid-FourRooms",
    entry_point="xminigrid.envs.minigrid.fourrooms:FourRooms",
)

# LockedRoom
register(
    id="MiniGrid-LockedRoom",
    entry_point="xminigrid.envs.minigrid.lockedroom:LockedRoom",
)

# Memory
register(
    id="MiniGrid-MemoryS8",
    entry_point="xminigrid.envs.minigrid.memory:Memory",
    width=8,
)

register(
    id="MiniGrid-MemoryS16",
    entry_point="xminigrid.envs.minigrid.memory:Memory",
    width=16,
)

register(
    id="MiniGrid-MemoryS32",
    entry_point="xminigrid.envs.minigrid.memory:Memory",
    width=32,
)

register(
    id="MiniGrid-MemoryS64",
    entry_point="xminigrid.envs.minigrid.memory:Memory",
    width=64,
)

register(
    id="MiniGrid-MemoryS128",
    entry_point="xminigrid.envs.minigrid.memory:Memory",
    width=128,
)

# PlayGround
register(
    id="MiniGrid-Playground",
    entry_point="xminigrid.envs.minigrid.playground:Playground",
)

# Unlock
register(
    id="MiniGrid-Unlock",
    entry_point="xminigrid.envs.minigrid.unlock:Unlock",
)

# UnlockPickUp
register(
    id="MiniGrid-UnlockPickUp",
    entry_point="xminigrid.envs.minigrid.unlockpickup:UnlockPickUp",
)



from __future__ import annotations

import abc
from typing import Union

import jax
import jax.numpy as jnp
from flax import struct

from ..types import AgentState, GridState
from .grid import equal, get_neighbouring_tiles, pad_along_axis

MAX_GOAL_ENCODING_LEN = 4 + 1  # for idx
NUM_GOALS = 15


def check_goal(
    encoding: jax.Array, grid: GridState, agent: AgentState, action: Union[int, jax.Array], position: jax.Array
) -> jax.Array:
    check = jax.lax.switch(
        encoding[0],
        (
            # empty goal first, we use them as paddings during benchmark generation
            lambda: EmptyGoal.decode(encoding)(grid, agent, action, position),
            lambda: AgentHoldGoal.decode(encoding)(grid, agent, action, position),
            lambda: AgentOnTileGoal.decode(encoding)(grid, agent, action, position),
            lambda: AgentNearGoal.decode(encoding)(grid, agent, action, position),
            lambda: TileNearGoal.decode(encoding)(grid, agent, action, position),
            lambda: TileOnPositionGoal.decode(encoding)(grid, agent, action, position),
            lambda: AgentOnPositionGoal.decode(encoding)(grid, agent, action, position),
            # goals for the extended benchmarks
            lambda: TileNearUpGoal.decode(encoding)(grid, agent, action, position),
            lambda: TileNearRightGoal.decode(encoding)(grid, agent, action, position),
            lambda: TileNearDownGoal.decode(encoding)(grid, agent, action, position),
            lambda: TileNearLeftGoal.decode(encoding)(grid, agent, action, position),
            lambda: AgentNearUpGoal.decode(encoding)(grid, agent, action, position),
            lambda: AgentNearRightGoal.decode(encoding)(grid, agent, action, position),
            lambda: AgentNearDownGoal.decode(encoding)(grid, agent, action, position),
            lambda: AgentNearLeftGoal.decode(encoding)(grid, agent, action, position),
        ),
    )
    return check


# should I inherit from abc.ABC?
class BaseGoal(struct.PyTreeNode):
    @abc.abstractmethod
    def __call__(
        self, grid: GridState, agent: AgentState, action: Union[int, jax.Array], position: jax.Array
    ) -> jax.Array: ...

    @classmethod
    @abc.abstractmethod
    def decode(cls, encoding: jax.Array) -> BaseGoal: ...

    @abc.abstractmethod
    def encode(self) -> jax.Array: ...


class EmptyGoal(BaseGoal):
    def __call__(self, grid, agent, action, position):
        return jnp.asarray(False)

    @classmethod
    def decode(cls, encoding):
        return cls()

    def encode(self):
        return jnp.zeros(MAX_GOAL_ENCODING_LEN, dtype=jnp.uint8)


class AgentHoldGoal(BaseGoal):
    tile: jax.Array

    def __call__(self, grid, agent, action, position):
        check = jax.lax.select(jnp.equal(action, 3), equal(agent.pocket, self.tile), jnp.asarray(False))
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(1), self.tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class AgentOnTileGoal(BaseGoal):
    tile: jax.Array

    def __call__(self, grid, agent, action, position):
        check = jax.lax.select(
            jnp.equal(action, 0), equal(grid[agent.position[0], agent.position[1]], self.tile), jnp.asarray(False)
        )
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(2), self.tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class AgentNearGoal(BaseGoal):
    tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _check_fn():
            up, right, down, left = get_neighbouring_tiles(grid, agent.position[0], agent.position[1])
            check = equal(up, self.tile) | equal(right, self.tile) | equal(down, self.tile) | equal(left, self.tile)

            return check

        check = jax.lax.select(jnp.equal(action, 0), _check_fn(), jnp.asarray(False))
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(3), self.tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class TileNearGoal(BaseGoal):
    tile_a: jax.Array
    tile_b: jax.Array

    def __call__(self, grid, agent, action, position):
        tile_a = self.tile_a
        tile_b = self.tile_b

        def _check_fn():
            tile = grid[position[0], position[1]]
            up, right, down, left = get_neighbouring_tiles(grid, position[0], position[1])

            check = jax.lax.select(
                jnp.logical_or(equal(tile, tile_a), equal(tile, tile_b)),
                jax.lax.select(
                    equal(tile, tile_a),
                    equal(up, tile_b) | equal(right, tile_b) | equal(down, tile_b) | equal(left, tile_b),
                    equal(up, tile_a) | equal(right, tile_a) | equal(down, tile_a) | equal(left, tile_a),
                ),
                jnp.asarray(False),
            )
            return check

        check = jax.lax.select(jnp.equal(action, 4), _check_fn(), jnp.asarray(False))
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(4), self.tile_a, self.tile_b], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class TileOnPositionGoal(BaseGoal):
    tile: jax.Array
    position: jax.Array

    def __call__(self, grid, agent, action, position):
        check = jnp.array_equal(grid[self.position[0], self.position[1]], self.tile)
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3], position=encoding[3:5])

    def encode(self):
        # WARN: uint8 type means that max grid size currently is 255
        encoding = jnp.hstack([jnp.asarray(5), self.tile, self.position], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class AgentOnPositionGoal(BaseGoal):
    position: jax.Array

    def __call__(self, grid, agent, action, position):
        check = jnp.array_equal(agent.position, self.position)
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(position=encoding[1:3])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(6), self.position], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class TileNearUpGoal(BaseGoal):
    tile_a: jax.Array
    tile_b: jax.Array

    def __call__(self, grid, agent, action, position):
        y, x = position
        tile = grid[y, x]

        def _check_fn():
            up, _, down, _ = get_neighbouring_tiles(grid, y, x)
            check = jnp.logical_or(
                equal(tile, self.tile_b) & equal(down, self.tile_a), equal(tile, self.tile_a) & equal(up, self.tile_b)
            )
            return check

        check = jax.lax.select(
            jnp.equal(action, 4) & (equal(tile, self.tile_a) | equal(tile, self.tile_b)),
            _check_fn(),
            jnp.asarray(False),
        )
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(7), self.tile_a, self.tile_b])
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class TileNearRightGoal(BaseGoal):
    tile_a: jax.Array
    tile_b: jax.Array

    def __call__(self, grid, agent, action, position):
        y, x = position
        tile = grid[y, x]

        def _check_fn():
            _, right, _, left = get_neighbouring_tiles(grid, y, x)
            check = jnp.logical_or(
                equal(tile, self.tile_b) & equal(left, self.tile_a),
                equal(tile, self.tile_a) & equal(right, self.tile_b),
            )
            return check

        check = jax.lax.select(
            jnp.equal(action, 4) & (equal(tile, self.tile_a) | equal(tile, self.tile_b)),
            _check_fn(),
            jnp.asarray(False),
        )
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(8), self.tile_a, self.tile_b])
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class TileNearDownGoal(BaseGoal):
    tile_a: jax.Array
    tile_b: jax.Array

    def __call__(self, grid, agent, action, position):
        y, x = position
        tile = grid[y, x]

        def _check_fn():
            up, _, down, _ = get_neighbouring_tiles(grid, y, x)
            check = jnp.logical_or(
                equal(tile, self.tile_b) & equal(up, self.tile_a), equal(tile, self.tile_a) & equal(down, self.tile_b)
            )
            return check

        check = jax.lax.select(
            jnp.equal(action, 4) & (equal(tile, self.tile_a) | equal(tile, self.tile_b)),
            _check_fn(),
            jnp.asarray(False),
        )
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(9), self.tile_a, self.tile_b])
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class TileNearLeftGoal(BaseGoal):
    tile_a: jax.Array
    tile_b: jax.Array

    def __call__(self, grid, agent, action, position):
        y, x = position
        tile = grid[y, x]

        def _check_fn():
            _, right, _, left = get_neighbouring_tiles(grid, y, x)
            check = jnp.logical_or(
                equal(tile, self.tile_b) & equal(right, self.tile_a),
                equal(tile, self.tile_a) & equal(left, self.tile_b),
            )
            return check

        check = jax.lax.select(
            jnp.equal(action, 4) & (equal(tile, self.tile_a) | equal(tile, self.tile_b)),
            _check_fn(),
            jnp.asarray(False),
        )
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(10), self.tile_a, self.tile_b])
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class AgentNearUpGoal(BaseGoal):
    tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _check_fn():
            up, _, _, _ = get_neighbouring_tiles(grid, agent.position[0], agent.position[1])
            check = equal(up, self.tile)
            return check

        check = jax.lax.select(
            jnp.equal(action, 0) | jnp.equal(action, 4),
            _check_fn(),
            jnp.asarray(False),
        )
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(11), self.tile])
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class AgentNearRightGoal(BaseGoal):
    tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _check_fn():
            _, right, _, _ = get_neighbouring_tiles(grid, agent.position[0], agent.position[1])
            check = equal(right, self.tile)
            return check

        check = jax.lax.select(
            jnp.equal(action, 0) | jnp.equal(action, 4),
            _check_fn(),
            jnp.asarray(False),
        )
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(12), self.tile])
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class AgentNearDownGoal(BaseGoal):
    tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _check_fn():
            _, _, down, _ = get_neighbouring_tiles(grid, agent.position[0], agent.position[1])
            check = equal(down, self.tile)
            return check

        check = jax.lax.select(
            jnp.equal(action, 0) | jnp.equal(action, 4),
            _check_fn(),
            jnp.asarray(False),
        )
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(13), self.tile])
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)


class AgentNearLeftGoal(BaseGoal):
    tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _check_fn():
            _, _, _, left = get_neighbouring_tiles(grid, agent.position[0], agent.position[1])
            check = equal(left, self.tile)
            return check

        check = jax.lax.select(
            jnp.equal(action, 0) | jnp.equal(action, 4),
            _check_fn(),
            jnp.asarray(False),
        )
        return check

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(14), self.tile])
        return pad_along_axis(encoding, MAX_GOAL_ENCODING_LEN)



# Code ported from https://github.com/Farama-Foundation/Minigrid/blob/master/minigrid/utils/rendering.py
from __future__ import annotations

import math
from typing import Callable, Union

import numpy as np
from typing_extensions import TypeAlias

Color: TypeAlias = Union[tuple[int, int, int], int, np.ndarray]
Point: TypeAlias = tuple[float, float]  # | np.ndarray


def downsample(img: np.ndarray, factor: int) -> np.ndarray:
    assert img.shape[0] % factor == 0
    assert img.shape[1] % factor == 0

    img = img.reshape([img.shape[0] // factor, factor, img.shape[1] // factor, factor, 3])
    img = img.mean(axis=3)
    img = img.mean(axis=1)

    return img


def fill_coords(img: np.ndarray, fn: Callable, color: Color) -> np.ndarray:
    for y in range(img.shape[0]):
        for x in range(img.shape[1]):
            yf = (y + 0.5) / img.shape[0]
            xf = (x + 0.5) / img.shape[1]
            if fn(xf, yf):
                img[y, x] = color

    return img


def rotate_fn(fin: Callable, cx: float, cy: float, theta: float) -> Callable:
    def fout(x, y):
        x = x - cx
        y = y - cy

        x2 = cx + x * math.cos(-theta) - y * math.sin(-theta)
        y2 = cy + y * math.cos(-theta) + x * math.sin(-theta)

        return fin(x2, y2)

    return fout


def point_in_line(x0: float, y0: float, x1: float, y1: float, r: float) -> Callable:
    p0 = np.array([x0, y0], dtype=np.float32)
    p1 = np.array([x1, y1], dtype=np.float32)
    dir = p1 - p0
    dist = np.linalg.norm(dir)
    dir = dir / dist

    xmin = min(x0, x1) - r
    xmax = max(x0, x1) + r
    ymin = min(y0, y1) - r
    ymax = max(y0, y1) + r

    def fn(x, y):
        # Fast, early escape test
        if x < xmin or x > xmax or y < ymin or y > ymax:
            return False

        q = np.array([x, y])
        pq = q - p0

        # Closest point on line
        a = np.dot(pq, dir)
        a = np.clip(a, 0, dist)
        p = p0 + a * dir

        dist_to_line = np.linalg.norm(q - p)
        return dist_to_line <= r

    return fn


def point_in_circle(cx: float, cy: float, r: float) -> Callable:
    def fn(x, y):
        return (x - cx) * (x - cx) + (y - cy) * (y - cy) <= r * r

    return fn


def point_in_rect(xmin: float, xmax: float, ymin: float, ymax: float) -> Callable:
    def fn(x, y):
        return x >= xmin and x <= xmax and y >= ymin and y <= ymax

    return fn


def point_in_triangle(a: Point, b: Point, c: Point) -> Callable:
    a_ = np.array(a, dtype=np.float32)
    b_ = np.array(b, dtype=np.float32)
    c_ = np.array(c, dtype=np.float32)

    def fn(x, y):
        v0 = c_ - a_
        v1 = b_ - a_
        v2 = np.array((x, y)) - a_

        # Compute dot products
        dot00 = np.dot(v0, v0)
        dot01 = np.dot(v0, v1)
        dot02 = np.dot(v0, v2)
        dot11 = np.dot(v1, v1)
        dot12 = np.dot(v1, v2)

        # Compute barycentric coordinates
        inv_denom = 1 / (dot00 * dot11 - dot01 * dot01)
        u = (dot11 * dot02 - dot01 * dot12) * inv_denom
        v = (dot00 * dot12 - dot01 * dot02) * inv_denom

        # Check if point is in triangle
        return (u >= 0) and (v >= 0) and (u + v) < 1

    return fn


def point_in_hexagon(s: float) -> Callable:
    def fn(x, y):
        x = abs(x - 0.5)
        y = abs(y - 0.5)
        return y < 3**0.5 * min(s - x, s / 2)

    return fn


def highlight_img(img: np.ndarray, color: Color = (255, 255, 255), alpha: float = 0.30) -> None:
    blend_img = img + alpha * (np.array(color, dtype=np.uint8) - img)
    blend_img = blend_img.clip(0, 255).astype(np.uint8)
    img[:, :, :] = blend_img



from __future__ import annotations

import argparse
import os
from typing import Optional

import imageio.v3 as iio
import jax
import numpy as np
import pygame
import pygame.freetype
from pygame.event import Event

import xminigrid

from .environment import Environment, EnvParamsT
from .rendering.text_render import print_ruleset
from .types import EnvCarryT


class ManualControl:
    def __init__(
        self,
        env: Environment[EnvParamsT, EnvCarryT],
        env_params: EnvParamsT,
        agent_view: bool = False,
        save_video: bool = False,
        video_path: Optional[str] = None,
        video_format: str = ".mp4",
        video_fps: int = 8,
    ):
        self.env = env
        self.env_params = env_params
        self.agent_view = agent_view
        self.save_video = save_video
        self.video_path = video_path
        self.video_format = video_format
        self.video_fps = video_fps

        if self.save_video:
            self.frames = []

        self._reset = jax.jit(self.env.reset)
        self._step = jax.jit(self.env.step)
        self._key = jax.random.key(0)

        self.timestep = None

        self.render_size = None
        self.window = None
        self.clock = None
        self.screen_size = 640
        self.closed = False

    def render(self) -> None:
        assert self.timestep is not None

        if self.agent_view:
            img = self.timestep.observation
        else:
            img = self.env.render(self.env_params, self.timestep)

        if self.save_video:
            self.frames.append(img)

        # [h, w, c] -> [w, h, c]
        img = np.transpose(img, axes=(1, 0, 2))

        if self.render_size is None:
            self.render_size = img.shape[:2]

        if self.window is None:
            pygame.init()
            pygame.display.init()
            self.window = pygame.display.set_mode((self.screen_size, self.screen_size))
            pygame.display.set_caption("xland-minigrid")
        if self.clock is None:
            self.clock = pygame.time.Clock()
        surf = pygame.surfarray.make_surface(img)

        # Create background with mission description
        offset = surf.get_size()[0] * 0.1
        # offset = 32 if self.agent_pov else 64
        bg = pygame.Surface((int(surf.get_size()[0] + offset), int(surf.get_size()[1] + offset)))
        bg.convert()
        bg.fill((255, 255, 255))
        bg.blit(surf, (offset / 2, 0))

        bg = pygame.transform.smoothscale(bg, (self.screen_size, self.screen_size))

        font_size = 16
        text = "TODO: game goal encoding"
        font = pygame.freetype.SysFont(pygame.font.get_default_font(), font_size)
        text_rect = font.get_rect(text, size=font_size)
        text_rect.center = bg.get_rect().center
        text_rect.y = int(bg.get_height() - font_size * 1.5)
        font.render_to(bg, text_rect, text, size=font_size)

        self.window.blit(bg, (0, 0))
        pygame.event.pump()
        self.clock.tick(10)
        pygame.display.flip()

    def start(self) -> None:
        self.reset()

        """Start the window display with blocking event loop"""
        while not self.closed:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.close()
                    break
                if event.type == pygame.KEYDOWN:
                    event.key = pygame.key.name(int(event.key))
                    self.key_handler(event)

    def step(self, action: int) -> None:
        self.timestep = self._step(self.env_params, self.timestep, action)
        print(
            f"Step: {self.timestep.state.step_num} | ",
            f"StepType: {self.timestep.step_type} | ",
            f"Discount: {self.timestep.discount} | ",
            f"Reward: {self.timestep.reward}",
        )
        self.render()

        if self.timestep.last():
            self.reset()

    def reset(self) -> None:
        print("Reset!")
        self._key, reset_key = jax.random.split(self._key)

        self.timestep = self._reset(self.env_params, reset_key)
        self.render()
        print(
            f"Step: {self.timestep.state.step_num} |",
            f"StepType: {self.timestep.step_type} |",
            f"Discount: {self.timestep.discount} |",
            f"Reward: {self.timestep.reward}",
        )

    def key_handler(self, event: Event) -> None:
        key: str = event.key
        print("pressed", key)

        if key == "escape":
            self.close()
            return

        key_to_action = {
            "up": 0,
            "right": 1,
            "left": 2,
            "tab": 3,
            "left shift": 4,
            "space": 5,
        }
        if key in key_to_action.keys():
            action = key_to_action[key]
            self.step(action)
        else:
            print(key)

    def close(self) -> None:
        if self.window:
            pygame.quit()

        if self.save_video:
            assert self.video_path is not None
            save_path = os.path.join(self.video_path, f"manual_control_rollout{self.video_format}")
            if self.video_format == ".mp4":
                iio.imwrite(save_path, self.frames, format_hint=".mp4", fps=self.video_fps)
            elif self.video_format == ".gif":
                iio.imwrite(
                    save_path, self.frames[:-1], format_hint=".gif", duration=(1000 * 1 / self.video_fps), loop=10
                )
                # iio.imwrite(save_path, self.frames, format_hint=".gif", duration=(1000 * 1 / self.video_fps), loop=10)
            else:
                raise RuntimeError("Unknown video format! Should be one of ('.mp4', '.gif')")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--env-id", type=str, default="MiniGrid-Empty-5x5", choices=xminigrid.registered_environments())
    parser.add_argument("--benchmark-id", type=str, default="trivial-1m", choices=xminigrid.registered_benchmarks())
    parser.add_argument("--ruleset-id", type=int, default=0)
    parser.add_argument("--agent-view", action="store_true")
    parser.add_argument("--save-video", action="store_true")
    parser.add_argument("--video-path", type=str, default=".")
    parser.add_argument("--video-format", type=str, default=".mp4", choices=(".mp4", ".gif"))
    parser.add_argument("--video-fps", type=int, default=5)

    args = parser.parse_args()
    env, env_params = xminigrid.make(args.env_id)

    if args.agent_view:
        from xminigrid.experimental.img_obs import RGBImgObservationWrapper

        env = RGBImgObservationWrapper(env)

    if "XLand" in args.env_id:
        bench = xminigrid.load_benchmark(args.benchmark_id)
        ruleset = bench.get_ruleset(args.ruleset_id)

        env_params = env_params.replace(ruleset=ruleset)
        print_ruleset(ruleset)
        print()

    control = ManualControl(
        env=env,
        env_params=env_params,
        agent_view=args.agent_view,
        save_video=args.save_video,
        video_path=args.video_path,
        video_format=args.video_format,
        video_fps=args.video_fps,
    )
    control.start()



from __future__ import annotations

import jax
import jax.numpy as jnp
from flax import struct

from ...core.constants import Colors, Tiles
from ...core.goals import EmptyGoal
from ...core.grid import cartesian_product_1d, nine_rooms, sample_coordinates, sample_direction
from ...core.rules import EmptyRule
from ...environment import Environment, EnvParams
from ...types import AgentState, EnvCarry, State

_goal_encoding = EmptyGoal().encode()
_rule_encoding = EmptyRule().encode()[None, ...]

# colors like in the original minigrid
_allowed_colors = jnp.array(
    (
        Colors.RED,
        Colors.GREEN,
        Colors.BLUE,
        Colors.PURPLE,
        Colors.YELLOW,
        Colors.GREY,
    ),
    dtype=jnp.uint8,
)

_allowed_doors = cartesian_product_1d(
    jnp.array((Tiles.DOOR_CLOSED,), dtype=jnp.uint8),
    _allowed_colors,
)
_allowed_objects = cartesian_product_1d(
    jnp.array((Tiles.BALL, Tiles.SQUARE, Tiles.PYRAMID, Tiles.KEY, Tiles.STAR, Tiles.HEX, Tiles.GOAL), dtype=jnp.uint8),
    _allowed_colors,
)
# number of doors with 9 rooms
_total_doors = 12


class PlaygroundEnvParams(EnvParams):
    num_objects: int = struct.field(pytree_node=False, default=12)


class Playground(Environment[PlaygroundEnvParams, EnvCarry]):
    def default_params(self, **kwargs) -> PlaygroundEnvParams:
        params = PlaygroundEnvParams(height=19, width=19)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # formula directly taken from MiniGrid
            params = params.replace(max_steps=512)
        return params

    def _generate_problem(self, params: PlaygroundEnvParams, key: jax.Array) -> State[EnvCarry]:
        key, *keys = jax.random.split(key, num=6)

        grid = nine_rooms(params.height, params.width)
        roomW, roomH = params.width // 3, params.height // 3

        # assuming that rooms are square!
        door_coords = jax.random.randint(keys[0], shape=(_total_doors,), minval=1, maxval=roomW)
        doors = jax.random.choice(keys[1], _allowed_doors, shape=(_total_doors,))

        # adapted from minigrid playground
        door_idx = 0
        for i in range(0, 3):
            for j in range(0, 3):
                xL = i * roomW
                yT = j * roomH
                xR = xL + roomW
                yB = yT + roomH

                if i + 1 < 3:
                    grid = grid.at[yT + door_coords[door_idx], xR].set(doors[door_idx])
                    door_idx = door_idx + 1

                if j + 1 < 3:
                    grid = grid.at[yB, xL + door_coords[door_idx]].set(doors[door_idx])
                    door_idx = door_idx + 1

        objects_coords = sample_coordinates(keys[2], grid, num=params.num_objects + 1)
        objects = jax.random.choice(keys[3], _allowed_objects, shape=(params.num_objects,))
        for i in range(params.num_objects):
            grid = grid.at[objects_coords[i][0], objects_coords[i][1]].set(objects[i])

        agent = AgentState(position=objects_coords[-1], direction=sample_direction(keys[4]))

        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=_goal_encoding,
            rule_encoding=_rule_encoding,
            carry=EnvCarry(),
        )
        return state



from __future__ import annotations

import jax
import jax.numpy as jnp

from ...core.constants import TILES_REGISTRY, Colors, Tiles
from ...core.goals import AgentOnTileGoal
from ...core.grid import horizontal_line, room, sample_direction, vertical_line
from ...core.rules import EmptyRule
from ...environment import Environment, EnvParams
from ...types import AgentState, EnvCarry, State

# goals and rules are hardcoded for minigrid envs
_goal_encoding = AgentOnTileGoal(tile=TILES_REGISTRY[Tiles.GOAL, Colors.GREEN]).encode()
_rule_encoding = EmptyRule().encode()[None, ...]

_wall_tile = TILES_REGISTRY[Tiles.WALL, Colors.GREY]

_allowed_colors = jnp.array(
    (
        Colors.YELLOW,
        Colors.PURPLE,
        Colors.GREEN,
        Colors.GREY,
        Colors.BLUE,
        Colors.RED,
    )
)


class LockedRoom(Environment[EnvParams, EnvCarry]):
    def default_params(self, **kwargs) -> EnvParams:
        params = EnvParams(height=19, width=19)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # formula directly taken from MiniGrid
            params = params.replace(max_steps=10 * params.height)
        return params

    def _generate_problem(self, params: EnvParams, key: jax.Array) -> State[EnvCarry]:
        key, rooms_key, colors_key, objects_key, coords_key, agent_pos_key, agent_dir_key = jax.random.split(key, num=7)

        # set up rooms
        grid = room(params.height, params.width)
        grid = vertical_line(grid, params.width // 2 - 2, 0, params.height, _wall_tile)
        grid = vertical_line(grid, params.width // 2 + 2, 0, params.height, _wall_tile)

        for i in range(1, 3):
            grid = horizontal_line(grid, 0, i * (params.height // 3), params.width // 2 - 2, _wall_tile)
            grid = horizontal_line(
                grid, params.width // 2 + 2, i * (params.height // 3), params.width // 2 - 2, _wall_tile
            )

        # hardcoded, but easier and faster to sample
        room_corners = jnp.array(
            (
                (0, 0),
                (params.height // 3, 0),
                (2 * (params.height // 3), 0),
                (0, params.width // 2 + 2),
                (params.height // 3, params.width // 2 + 2),
                (2 * (params.height // 3), params.width // 2 + 2),
            )
        )
        doors_idxs = jnp.array(
            (
                # left doors
                (params.height // 2 - (params.height // 3), params.width // 2 - 2),
                (params.height // 2, params.width // 2 - 2),
                (params.height // 2 + (params.height // 3), params.width // 2 - 2),
                # right doors
                (params.height // 2 - (params.height // 3), params.width // 2 + 2),
                (params.height // 2, params.width // 2 + 2),
                (params.height // 2 + (params.height // 3), params.width // 2 + 2),
            )
        )
        colors_order = jax.random.permutation(colors_key, _allowed_colors)

        for i in range(6):
            grid = grid.at[doors_idxs[i][0], doors_idxs[i][1]].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, colors_order[i]])

        goal_idx, key_idx = jax.random.choice(rooms_key, 6, shape=(2,), replace=False)
        coords = jax.random.randint(
            coords_key,
            shape=(2, 2),
            minval=jnp.array((1, 1)),
            maxval=jnp.array((params.height // 3, params.width // 2 - 2)),
        )

        # set up goal room and locked door (I know, it's a lot of indexing)
        target_color = grid[doors_idxs[goal_idx][0], doors_idxs[goal_idx][1], 1]
        grid = grid.at[doors_idxs[goal_idx][0], doors_idxs[goal_idx][1]].set(
            TILES_REGISTRY[Tiles.DOOR_LOCKED, target_color]
        )
        grid = grid.at[room_corners[goal_idx][0] + coords[0][0], room_corners[goal_idx][1] + coords[0][1]].set(
            TILES_REGISTRY[Tiles.GOAL, Colors.GREEN]
        )
        # place key
        grid = grid.at[room_corners[key_idx][0] + coords[1][0], room_corners[key_idx][1] + coords[1][1]].set(
            TILES_REGISTRY[Tiles.KEY, target_color]
        )

        # sample agent position
        position = jax.random.randint(
            agent_pos_key,
            shape=(2,),
            minval=jnp.array((1, params.width // 2 - 1)),
            maxval=jnp.array((params.height - 1, params.width // 2 + 1)),
        )
        agent = AgentState(position=position, direction=sample_direction(agent_dir_key))

        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=_goal_encoding,
            rule_encoding=_rule_encoding,
            carry=EnvCarry(),
        )
        return state



from __future__ import annotations

import jax
import jax.numpy as jnp
from typing_extensions import TypeAlias

from ..types import AgentState, GridState, IntOrArray
from .constants import DIRECTIONS, TILES_REGISTRY, Colors, Tiles
from .grid import check_can_put, check_pickable, check_walkable, equal

ActionOutput: TypeAlias = tuple[GridState, AgentState, jax.Array]


def _move(position: jax.Array, direction: jax.Array) -> jax.Array:
    direction = jax.lax.dynamic_index_in_dim(DIRECTIONS, direction, keepdims=False)
    new_position = position + direction
    return new_position


def move_forward(grid: GridState, agent: AgentState) -> ActionOutput:
    next_position = jnp.clip(
        _move(agent.position, agent.direction),
        min=jnp.array((0, 0)),
        max=jnp.array((grid.shape[0] - 1, grid.shape[1] - 1)),  # H, W
    )
    position = jax.lax.select(
        check_walkable(grid, next_position),
        next_position,
        agent.position,
    )
    new_agent = agent.replace(position=position)

    return grid, new_agent, new_agent.position


def turn_clockwise(grid: GridState, agent: AgentState) -> ActionOutput:
    new_direction = (agent.direction + 1) % 4
    new_agent = agent.replace(direction=new_direction)
    return grid, new_agent, agent.position


def turn_counterclockwise(grid: GridState, agent: AgentState) -> ActionOutput:
    new_direction = (agent.direction - 1) % 4
    new_agent = agent.replace(direction=new_direction)
    return grid, new_agent, agent.position


def pick_up(grid: GridState, agent: AgentState) -> ActionOutput:
    next_position = _move(agent.position, agent.direction)

    is_pickable = check_pickable(grid, next_position)
    is_empty_pocket = jnp.equal(agent.pocket[0], Tiles.EMPTY)
    # pick up only if pocket is empty and entity is pickable
    new_grid, new_agent = jax.lax.cond(
        is_pickable & is_empty_pocket,
        lambda: (
            grid.at[next_position[0], next_position[1]].set(TILES_REGISTRY[Tiles.FLOOR, Colors.BLACK]),
            agent.replace(
                pocket=TILES_REGISTRY[
                    grid[next_position[0], next_position[1], 0],
                    grid[next_position[0], next_position[1], 1],
                ]
            ),
        ),
        lambda: (grid, agent),
    )
    return new_grid, new_agent, next_position


def put_down(grid: GridState, agent: AgentState) -> ActionOutput:
    next_position = _move(agent.position, agent.direction)

    can_put = check_can_put(grid, next_position)
    not_empty_pocket = jnp.not_equal(agent.pocket[0], Tiles.EMPTY)
    new_grid, new_agent = jax.lax.cond(
        can_put & not_empty_pocket,
        lambda: (
            grid.at[next_position[0], next_position[1]].set(agent.pocket),
            agent.replace(pocket=TILES_REGISTRY[Tiles.EMPTY, Colors.EMPTY]),
        ),
        lambda: (grid, agent),
    )
    return new_grid, new_agent, next_position


# TODO: may be this should be open_door action? toggle is too general and box is not supported yet
def toggle(grid: GridState, agent: AgentState) -> ActionOutput:
    next_position = _move(agent.position, agent.direction)
    next_tile = grid[next_position[0], next_position[1]]

    # check door_locked
    new_grid = jax.lax.select(
        jnp.equal(next_tile[0], Tiles.DOOR_LOCKED) & equal(agent.pocket, TILES_REGISTRY[Tiles.KEY, next_tile[1]]),
        grid.at[next_position[0], next_position[1]].set(TILES_REGISTRY[Tiles.DOOR_OPEN, next_tile[1]]),
        grid,
    )
    # check door_closed
    new_grid = jax.lax.select(
        jnp.equal(next_tile[0], Tiles.DOOR_CLOSED),
        grid.at[next_position[0], next_position[1]].set(TILES_REGISTRY[Tiles.DOOR_OPEN, next_tile[1]]),
        new_grid,
    )
    # check door_open
    new_grid = jax.lax.select(
        jnp.equal(next_tile[0], Tiles.DOOR_OPEN),
        grid.at[next_position[0], next_position[1]].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, next_tile[1]]),
        new_grid,
    )
    return new_grid, agent, next_position


def take_action(grid: GridState, agent: AgentState, action: IntOrArray) -> ActionOutput:
    # This will evaluate all actions.
    # Can we fix this and choose only one function? It'll speed everything up dramatically.
    actions = (
        lambda: move_forward(grid, agent),
        lambda: turn_clockwise(grid, agent),
        lambda: turn_counterclockwise(grid, agent),
        lambda: pick_up(grid, agent),
        lambda: put_down(grid, agent),
        lambda: toggle(grid, agent),
    )
    new_grid, new_agent, changed_position = jax.lax.switch(action, actions)

    return new_grid, new_agent, changed_position



from __future__ import annotations

import jax
import jax.numpy as jnp

from ...core.constants import TILES_REGISTRY, Colors, Tiles
from ...core.goals import AgentOnTileGoal
from ...core.grid import coordinates_mask, room, sample_coordinates, sample_direction, vertical_line
from ...core.rules import EmptyRule
from ...environment import Environment, EnvParams
from ...types import AgentState, EnvCarry, State

_goal_encoding = AgentOnTileGoal(tile=TILES_REGISTRY[Tiles.GOAL, Colors.GREEN]).encode()
_rule_encoding = EmptyRule().encode()[None, ...]


class DoorKey(Environment[EnvParams, EnvCarry]):
    def default_params(self, **kwargs) -> EnvParams:
        params = EnvParams(height=5, width=5)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # formula directly taken from MiniGrid
            params = params.replace(max_steps=10 * (params.height * params.width))
        return params

    def _generate_problem(self, params: EnvParams, key: jax.Array) -> State[EnvCarry]:
        key, _key = jax.random.split(key)
        keys = jax.random.split(_key, num=4)

        wall_pos = jax.random.randint(keys[0], shape=(), minval=2, maxval=params.width - 2)
        door_pos = jax.random.randint(keys[1], shape=(), minval=1, maxval=params.width - 2)

        grid = room(params.height, params.width)
        grid = vertical_line(grid, wall_pos, 0, params.height, tile=TILES_REGISTRY[Tiles.WALL, Colors.GREY])
        grid = grid.at[door_pos, wall_pos].set(TILES_REGISTRY[Tiles.DOOR_LOCKED, Colors.YELLOW])
        grid = grid.at[params.height - 2, params.width - 2].set(TILES_REGISTRY[Tiles.GOAL, Colors.GREEN])

        # mask out positions after the wall, so that agent and key are always on the opposite side from the goal
        # WARN: this is a bit expensive, judging by the FPS benchmark
        mask = coordinates_mask(grid, (params.height, wall_pos), comparison_fn=jnp.less)
        key_coords, agent_coords = sample_coordinates(keys[2], grid, num=2, mask=mask)

        grid = grid.at[key_coords[0], key_coords[1]].set(TILES_REGISTRY[Tiles.KEY, Colors.YELLOW])

        agent = AgentState(position=agent_coords, direction=sample_direction(keys[3]))
        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=_goal_encoding,
            rule_encoding=_rule_encoding,
            carry=EnvCarry(),
        )
        return state



# jit-compatible RGB observations. Currently experimental!
# if it proves useful and necessary in the future, I will consider rewriting env.render in such style also
from __future__ import annotations

import os

import jax
import jax.numpy as jnp
import numpy as np

from ..benchmarks import load_bz2_pickle, save_bz2_pickle
from ..core.constants import NUM_COLORS, NUM_LAYERS, TILES_REGISTRY
from ..rendering.rgb_render import render_tile
from ..wrappers import Wrapper

CACHE_PATH = os.environ.get("XLAND_MINIGRID_CACHE", os.path.expanduser("~/.xland_minigrid"))
FORCE_RELOAD = os.environ.get("XLAND_MINIGRID_RELOAD_CACHE", False)


def build_cache(tiles: np.ndarray, tile_size: int = 32) -> tuple[np.ndarray, np.ndarray]:
    cache = np.zeros((tiles.shape[0], tiles.shape[1], tile_size, tile_size, 3), dtype=np.uint8)
    agent_cache = np.zeros((tiles.shape[0], tiles.shape[1], tile_size, tile_size, 3), dtype=np.uint8)

    for y in range(tiles.shape[0]):
        for x in range(tiles.shape[1]):
            # rendering tile
            tile_img = render_tile(
                tile=tuple(tiles[y, x]),
                agent_direction=None,
                highlight=False,
                tile_size=int(tile_size),
            )
            cache[y, x] = tile_img

            # rendering agent on top
            tile_w_agent_img = render_tile(
                tile=tuple(tiles[y, x]),
                agent_direction=0,
                highlight=False,
                tile_size=int(tile_size),
            )
            agent_cache[y, x] = tile_w_agent_img

    return cache, agent_cache


# building cache of pre-rendered tiles
TILE_SIZE = 32

cache_path = os.path.join(CACHE_PATH, "render_cache")

if not os.path.exists(cache_path) or FORCE_RELOAD:
    os.makedirs(CACHE_PATH, exist_ok=True)
    print("Building rendering cache, may take a while...")
    TILE_CACHE, TILE_W_AGENT_CACHE = build_cache(np.asarray(TILES_REGISTRY), tile_size=TILE_SIZE)
    TILE_CACHE = jnp.asarray(TILE_CACHE).reshape(-1, TILE_SIZE, TILE_SIZE, 3)
    TILE_W_AGENT_CACHE = jnp.asarray(TILE_W_AGENT_CACHE).reshape(-1, TILE_SIZE, TILE_SIZE, 3)

    print(f"Done. Cache is saved to {cache_path} and will be reused on consequent runs.")
    save_bz2_pickle({"tile_cache": TILE_CACHE, "tile_agent_cache": TILE_W_AGENT_CACHE}, cache_path)

TILE_CACHE = load_bz2_pickle(cache_path)["tile_cache"]
TILE_W_AGENT_CACHE = load_bz2_pickle(cache_path)["tile_agent_cache"]


# rendering with cached tiles
def _render_obs(obs: jax.Array) -> jax.Array:
    view_size = obs.shape[0]

    obs_flat_idxs = obs[:, :, 0] * NUM_COLORS + obs[:, :, 1]
    # render all tiles
    rendered_obs = jnp.take(TILE_CACHE, obs_flat_idxs, axis=0)

    # add agent tile
    agent_tile = TILE_W_AGENT_CACHE[obs_flat_idxs[view_size - 1, view_size // 2]]
    rendered_obs = rendered_obs.at[view_size - 1, view_size // 2].set(agent_tile)
    # [view_size, view_size, tile_size, tile_size, 3] -> [view_size * tile_size, view_size * tile_size, 3]
    rendered_obs = rendered_obs.transpose((0, 2, 1, 3, 4)).reshape(view_size * TILE_SIZE, view_size * TILE_SIZE, 3)

    return rendered_obs


class RGBImgObservationWrapper(Wrapper):
    def observation_shape(self, params):
        new_shape = (params.view_size * TILE_SIZE, params.view_size * TILE_SIZE, 3)

        base_shape = self._env.observation_shape(params)
        if isinstance(base_shape, dict):
            assert "img" in base_shape
            obs_shape = {**base_shape, **{"img": new_shape}}
        else:
            obs_shape = new_shape

        return obs_shape

    def __convert_obs(self, timestep):
        if isinstance(timestep.observation, dict):
            assert "img" in timestep.observation
            rendered_obs = {**timestep.observation, **{"img": _render_obs(timestep.observation["img"])}}
        else:
            rendered_obs = _render_obs(timestep.observation)

        timestep = timestep.replace(observation=rendered_obs)
        return timestep

    def reset(self, params, key):
        timestep = self._env.reset(params, key)
        timestep = self.__convert_obs(timestep)
        return timestep

    def step(self, params, timestep, action):
        timestep = self._env.step(params, timestep, action)
        timestep = self.__convert_obs(timestep)
        return timestep



from __future__ import annotations

from typing import Any, Union

import jax

from .environment import Environment, EnvParamsT
from .types import EnvCarryT, IntOrArray, State, TimeStep


class Wrapper(Environment[EnvParamsT, EnvCarryT]):
    def __init__(self, env: Environment[EnvParamsT, EnvCarryT]):
        self._env = env

    # Question: what if wrapper adds new parameters to the dataclass?
    # Solution: do this after applying the wrapper:
    #   env_params = wrapped_env.default_params(**dataclasses.asdict(original_params))
    def default_params(self, **kwargs) -> EnvParamsT:
        return self._env.default_params(**kwargs)

    def num_actions(self, params: EnvParamsT) -> int:
        return self._env.num_actions(params)

    def observation_shape(self, params: EnvParamsT) -> Union[tuple[int, int, int], dict[str, Any]]:
        return self._env.observation_shape(params)

    def _generate_problem(self, params: EnvParamsT, key: jax.Array) -> State[EnvCarryT]:
        return self._env._generate_problem(params, key)

    def reset(self, params: EnvParamsT, key: jax.Array) -> TimeStep[EnvCarryT]:
        return self._env.reset(params, key)

    def step(self, params: EnvParamsT, timestep: TimeStep[EnvCarryT], action: IntOrArray) -> TimeStep[EnvCarryT]:
        return self._env.step(params, timestep, action)

    def render(self, params: EnvParamsT, timestep: TimeStep[EnvCarryT]):
        return self._env.render(params, timestep)


# gym and gymnasium style reset (on the same step with termination)
class GymAutoResetWrapper(Wrapper):
    def __auto_reset(self, params, timestep):
        key, _ = jax.random.split(timestep.state.key)
        reset_timestep = self._env.reset(params, key)

        timestep = timestep.replace(
            state=reset_timestep.state,
            observation=reset_timestep.observation,
        )
        return timestep

    # TODO: add last_obs somewhere in the timestep? add extras like in Jumanji?
    def step(self, params, timestep, action):
        timestep = self._env.step(params, timestep, action)
        timestep = jax.lax.cond(
            timestep.last(),
            lambda: self.__auto_reset(params, timestep),
            lambda: timestep,
        )
        return timestep


# dm_env and envpool style reset (on the next step after termination)
class DmEnvAutoResetWrapper(Wrapper):
    def step(self, params, timestep, action):
        timestep = jax.lax.cond(
            timestep.last(),
            lambda: self._env.reset(params, timestep.state.key),
            lambda: self._env.step(params, timestep, action),
        )
        return timestep


# Yes, these are a bit stupid, but a tmp workaround to not write an actual system for spaces.
# May be, in the future, I will port the entire API to some existing one, like functional Gymnasium.
# For now, faster to do this stuff with dicts instead...
# NB: if you do not want to use this (due to the dicts as obs),
# just get needed parts from the original TimeStep and State dataclasses
class DirectionObservationWrapper(Wrapper):
    def observation_shape(self, params):
        base_shape = self._env.observation_shape(params)
        if isinstance(base_shape, dict):
            assert "img" in base_shape
            obs_shape = {**base_shape, **{"direction": 4}}
        else:
            obs_shape = {
                "img": self._env.observation_shape(params),
                "direction": 4,
            }
        return obs_shape

    def __extend_obs(self, timestep):
        direction = jax.nn.one_hot(timestep.state.agent.direction, num_classes=4)
        if isinstance(timestep.observation, dict):
            assert "img" in timestep.observation
            extended_obs = {
                **timestep.observation,
                **{"direction": direction},
            }
        else:
            extended_obs = {
                "img": timestep.observation,
                "direction": direction,
            }

        timestep = timestep.replace(observation=extended_obs)
        return timestep

    def reset(self, params, key):
        timestep = self._env.reset(params, key)
        timestep = self.__extend_obs(timestep)
        return timestep

    def step(self, params, timestep, action):
        timestep = self._env.step(params, timestep, action)
        timestep = self.__extend_obs(timestep)
        return timestep


class RulesAndGoalsObservationWrapper(Wrapper):
    def observation_shape(self, params):
        base_shape = self._env.observation_shape(params)
        if isinstance(base_shape, dict):
            assert "img" in base_shape
            obs_shape = {
                **base_shape,
                **{
                    "goal_encoding": params.ruleset.goal.shape,
                    "rule_encoding": params.ruleset.rules.shape,
                },
            }
        else:
            obs_shape = {
                "img": self._env.observation_shape(params),
                "goal_encoding": params.ruleset.goal.shape,
                "rule_encoding": params.ruleset.rules.shape,
            }
        return obs_shape

    def __extend_obs(self, timestep):
        goal_encoding = timestep.state.goal_encoding
        rule_encoding = timestep.state.rule_encoding
        if isinstance(timestep.observation, dict):
            assert "img" in timestep.observation
            extended_obs = {
                **timestep.observation,
                **{
                    "goal_encoding": goal_encoding,
                    "rule_encoding": rule_encoding,
                },
            }
        else:
            extended_obs = {
                "img": timestep.observation,
                "goal_encoding": goal_encoding,
                "rule_encoding": rule_encoding,
            }

        timestep = timestep.replace(observation=extended_obs)
        return timestep

    def reset(self, params, key):
        timestep = self._env.reset(params, key)
        timestep = self.__extend_obs(timestep)
        return timestep

    def step(self, params, timestep, action):
        timestep = self._env.step(params, timestep, action)
        timestep = self.__extend_obs(timestep)
        return timestep



import jax
import jax.numpy as jnp

from ..types import AgentState, GridState
from .constants import Tiles
from .grid import align_with_up, check_see_behind


def crop_field_of_view(grid: GridState, agent: AgentState, height: int, width: int) -> jax.Array:
    # TODO: assert height and width are odd and >= 3
    # TODO: in theory we don't need padding from all 4 sides, only for out of bounds sides
    grid = jnp.pad(
        grid,
        pad_width=((height, height), (width, width), (0, 0)),
        constant_values=Tiles.EMPTY,
    )
    # account for padding
    y = agent.position[0] + height
    x = agent.position[1] + width

    # to make slice_size static we compute top left corner of fov rectangle, then slice with static H, W
    start_indices = jax.lax.switch(
        agent.direction,
        (
            lambda: (y - height + 1, x - (width // 2), 0),
            lambda: (y - (height // 2), x, 0),
            lambda: (y, x - (width // 2), 0),
            lambda: (y - (height // 2), x - width + 1, 0),
        ),
    )
    fov_crop = jax.lax.dynamic_slice(grid, start_indices, (height, width, 2))

    return fov_crop


def transparent_field_of_view(grid: GridState, agent: AgentState, height: int, width: int) -> jax.Array:
    fov_grid = crop_field_of_view(grid, agent, height, width)
    fov_grid = align_with_up(fov_grid, agent.direction)

    # TODO: should we even do this? Agent with good memory can remember what he picked up.
    # WARN: this can overwrite tile the agent is on, GOAL for example.
    # https://github.com/Farama-Foundation/Minigrid/blob/463f07f20ff5cc2843d4b7cdf12c54a73b7d4d1c/minigrid/minigrid_env.py#L618 # noqa
    # Make it so the agent sees what it's carrying
    # We do this by placing the carried object at the agent's position
    # in the agent's partially observable view
    # fov_grid = fov_grid.at[height - 1, width // 2].set(agent.pocket)

    return fov_grid


# Direct port from original minigrid:
# https://github.com/Farama-Foundation/Minigrid/blob/e6f34bee70c5eb45ca9bfa2ea061cf06dd03e7b3/minigrid/core/grid.py#L291C9-L291C20 # noqa
# but adapted to jax and transposed grid
# WARN: only works for field of view crop aligned with UP direction, use align_with_up before!
def generate_viz_mask_minigrid(grid: GridState) -> jax.Array:
    H, W = grid.shape[0], grid.shape[1]
    viz_mask = jnp.zeros((H, W), dtype=jnp.bool_)
    # agent position with UP alignment, always visible
    viz_mask = viz_mask.at[H - 1, W // 2].set(True)

    def _forward_set_visible(mask, y, x):
        mask = mask.at[y, x + 1].set(True)
        mask = jax.lax.select(
            y > 0,
            mask.at[y - 1, x + 1].set(True).at[y - 1, x].set(True),
            mask,
        )
        return mask

    def _backward_set_visible(mask, y, x):
        mask = mask.at[y, x - 1].set(True)
        mask = jax.lax.select(
            y > 0,
            mask.at[y - 1, x - 1].set(True).at[y - 1, x].set(True),
            mask,
        )
        return mask

    # TODO: precompute check_see_behind mask
    def _forward_body(carry, x):
        viz_mask, y = carry
        viz_mask = jax.lax.select(
            jnp.logical_and(viz_mask[y, x], check_see_behind(grid, jnp.array((y, x)))),
            _forward_set_visible(viz_mask, y, x),
            viz_mask,
        )
        return (viz_mask, y), None

    def _backward_body(carry, x):
        viz_mask, y = carry
        viz_mask = jax.lax.select(
            jnp.logical_and(viz_mask[y, x], check_see_behind(grid, jnp.array((y, x)))),
            _backward_set_visible(viz_mask, y, x),
            viz_mask,
        )
        return (viz_mask, y), None

    def _main_body(viz_mask, y):
        (viz_mask, _), _ = jax.lax.scan(f=_forward_body, init=(viz_mask, y), xs=jnp.arange(0, W - 1))
        (viz_mask, _), _ = jax.lax.scan(f=_backward_body, init=(viz_mask, y), xs=jnp.arange(1, W), reverse=True)
        return viz_mask, None

    viz_mask, _ = jax.lax.scan(f=_main_body, init=(viz_mask), xs=jnp.arange(0, H), reverse=True)

    return viz_mask


# TODO: works well with unroll=16 and random actions, but very slow with PPO even with high unroll!
def minigrid_field_of_view(grid: GridState, agent: AgentState, height: int, width: int) -> jax.Array:
    fov_grid = crop_field_of_view(grid, agent, height, width)
    fov_grid = align_with_up(fov_grid, agent.direction)
    mask = generate_viz_mask_minigrid(fov_grid)
    # set EMPTY as unseen value for all layers (including colors, as EMPTY color has same id value)
    fov_grid = jnp.where(mask[..., None], fov_grid, Tiles.EMPTY)

    # TODO: should we even do this? Agent with good memory can remember what he picked up.
    # WARN: this can overwrite tile the agent is on, GOAL for example.
    # https://github.com/Farama-Foundation/Minigrid/blob/463f07f20ff5cc2843d4b7cdf12c54a73b7d4d1c/minigrid/minigrid_env.py#L618 # noqa
    # Make it so the agent sees what it's carrying
    # We do this by placing the carried object at the agent's position
    # in the agent's partially observable view
    # fov_grid = fov_grid.at[height - 1, width // 2].set(agent.pocket)

    return fov_grid



from typing import Optional, Union

import jax
import jax.numpy as jnp

from ..core.constants import Colors, Tiles
from ..types import AgentState, RuleSet

COLOR_NAMES = {
    Colors.EMPTY: "white",
    Colors.RED: "red",
    Colors.GREEN: "green",
    Colors.BLUE: "blue",
    Colors.PURPLE: "purple",
    Colors.YELLOW: "yellow",
    Colors.GREY: "grey",
    Colors.BLACK: "black",
    Colors.ORANGE: "orange",
    Colors.WHITE: "white",
    Colors.BROWN: "brown",
    Colors.PINK: "pink",
}

TILE_STR = {
    Tiles.EMPTY: " ",
    Tiles.FLOOR: ".",
    Tiles.WALL: "☰",
    Tiles.BALL: "⏺",
    Tiles.SQUARE: "▪",
    Tiles.PYRAMID: "▲",
    Tiles.HEX: "⬢",
    Tiles.STAR: "★",
    Tiles.GOAL: "■",
    Tiles.DOOR_LOCKED: "L",
    Tiles.DOOR_CLOSED: "C",
    Tiles.DOOR_OPEN: "O",
    Tiles.KEY: "K",
}

# for ruleset printing
RULE_TILE_STR = {
    Tiles.FLOOR: "floor",
    Tiles.BALL: "ball",
    Tiles.SQUARE: "square",
    Tiles.PYRAMID: "pyramid",
    Tiles.GOAL: "goal",
    Tiles.KEY: "key",
    Tiles.HEX: "hexagon",
    Tiles.STAR: "star",
}

PLAYER_STR = {0: "^", 1: ">", 2: "V", 3: "<"}


def _wrap_with_color(string: str, color: str) -> str:
    return f"[bold {color}]{string}[/bold {color}]"


# WARN: will NOT work under jit and needed for debugging mainly.
def render(grid: jax.Array, agent: Optional[AgentState] = None) -> str:
    string = ""

    for y in range(grid.shape[0]):
        for x in range(grid.shape[1]):
            tile_id, tile_color = grid[y, x]
            tile_str = TILE_STR[tile_id.item()]
            tile_color = COLOR_NAMES[tile_color.item()]

            if agent is not None and jnp.all(agent.position == jnp.array((y, x))):
                tile_str = PLAYER_STR[agent.direction.item()]
                tile_color = COLOR_NAMES[Colors.RED]

            string += _wrap_with_color(tile_str, tile_color)

        if y < grid.shape[0] - 1:
            string += "\n"

    return string


# WARN: This is for debugging mainly! Will refactor later if needed.
def _encode_tile(tile: list[int]) -> str:
    return f"{COLOR_NAMES[tile[1]]} {RULE_TILE_STR[tile[0]]}"


def _text_encode_goal(goal: list[int]) -> str:
    goal_id = goal[0]
    if goal_id == 1:
        return f"AgentHold({_encode_tile(goal[1:3])})"
    elif goal_id == 3:
        return f"AgentNear({_encode_tile(goal[1:3])})"
    elif goal_id == 4:
        return f"TileNear({_encode_tile(goal[1:3])}, {_encode_tile(goal[3:5])})"
    elif goal_id == 7:
        return f"TileNearUpGoal({_encode_tile(goal[1:3])}, {_encode_tile(goal[3:5])})"
    elif goal_id == 8:
        return f"TileNearRightGoal({_encode_tile(goal[1:3])}, {_encode_tile(goal[3:5])})"
    elif goal_id == 9:
        return f"TileNearDownGoal({_encode_tile(goal[1:3])}, {_encode_tile(goal[3:5])})"
    elif goal_id == 10:
        return f"TileNearLeftGoal({_encode_tile(goal[1:3])}, {_encode_tile(goal[3:5])})"
    elif goal_id == 11:
        return f"AgentNearUpGoal({_encode_tile(goal[1:3])})"
    elif goal_id == 12:
        return f"AgentNearRightGoal({_encode_tile(goal[1:3])})"
    elif goal_id == 13:
        return f"AgentNearDownGoal({_encode_tile(goal[1:3])})"
    elif goal_id == 14:
        return f"AgentNearLeftGoal({_encode_tile(goal[1:3])})"
    else:
        raise RuntimeError(f"Rendering: Unknown goal id: {goal_id}")


def _text_encode_rule(rule: list[int]) -> str:
    rule_id = rule[0]
    if rule_id == 1:
        return f"AgentHold({_encode_tile(rule[1:3])}) -> {_encode_tile(rule[3:5])}"
    elif rule_id == 2:
        return f"AgentNear({_encode_tile(rule[1:3])}) -> {_encode_tile(rule[3:5])}"
    elif rule_id == 3:
        return f"TileNear({_encode_tile(rule[1:3])}, {_encode_tile(rule[3:5])}) -> {_encode_tile(rule[5:7])}"
    elif rule_id == 4:
        return f"TileNearUpRule({_encode_tile(rule[1:3])}, {_encode_tile(rule[3:5])}) -> {_encode_tile(rule[5:7])}"
    elif rule_id == 5:
        return f"TileNearRightRule({_encode_tile(rule[1:3])}, {_encode_tile(rule[3:5])}) -> {_encode_tile(rule[5:7])}"
    elif rule_id == 6:
        return f"TileNearDownRule({_encode_tile(rule[1:3])}, {_encode_tile(rule[3:5])}) -> {_encode_tile(rule[5:7])}"
    elif rule_id == 7:
        return f"TileNearLeftRule({_encode_tile(rule[1:3])}, {_encode_tile(rule[3:5])}) -> {_encode_tile(rule[5:7])}"
    elif rule_id == 8:
        return f"AgentNearUpRule({_encode_tile(rule[1:3])}) -> {_encode_tile(rule[3:5])}"
    elif rule_id == 9:
        return f"AgentNearRightRule({_encode_tile(rule[1:3])}) -> {_encode_tile(rule[3:5])}"
    elif rule_id == 10:
        return f"AgentNearDownRule({_encode_tile(rule[1:3])}) -> {_encode_tile(rule[3:5])}"
    elif rule_id == 11:
        return f"AgentNearLeftRule({_encode_tile(rule[1:3])}) -> {_encode_tile(rule[3:5])}"
    else:
        raise RuntimeError(f"Rendering: Unknown rule id: {rule_id}")


def print_ruleset(ruleset: RuleSet) -> None:
    print("GOAL:")
    print(_text_encode_goal(ruleset.goal.tolist()))
    print()
    print("RULES:")
    for rule in ruleset.rules.tolist():
        if rule[0] != 0:
            print(_text_encode_rule(rule))
    print()
    print("INIT TILES:")
    for tile in ruleset.init_tiles.tolist():
        if tile[0] != 0:
            print(_encode_tile(tile))



from __future__ import annotations

import jax
import jax.numpy as jnp

from ...core.constants import TILES_REGISTRY, Colors, Tiles
from ...core.goals import AgentHoldGoal
from ...core.grid import coordinates_mask, sample_coordinates, sample_direction, two_rooms
from ...core.rules import EmptyRule
from ...environment import Environment, EnvParams
from ...types import AgentState, EnvCarry, State

# colors like in the original minigrid
_allowed_colors = jnp.array(
    (
        Colors.RED,
        Colors.GREEN,
        Colors.BLUE,
        Colors.PURPLE,
        Colors.YELLOW,
        Colors.GREY,
    )
)
_allowed_entities = jnp.array(
    (
        Tiles.BALL,
        Tiles.SQUARE,
        Tiles.PYRAMID,
        Tiles.HEX,
        Tiles.STAR,
    )
)
_rule_encoding = EmptyRule().encode()[None, ...]


class BlockedUnlockPickUp(Environment[EnvParams, EnvCarry]):
    def default_params(self, **kwargs) -> EnvParams:
        params = EnvParams(height=6, width=11)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # formula directly taken from MiniGrid
            params = params.replace(max_steps=16 * params.height**2)
        return params

    def _generate_problem(self, params: EnvParams, key: jax.Array) -> State[EnvCarry]:
        key, _key = jax.random.split(key)
        keys = jax.random.split(_key, num=7)

        pickup_obj, block_obj = jax.random.choice(keys[0], _allowed_entities, shape=(2,))
        door_color, pickup_obj_color, block_obj_color = jax.random.choice(keys[1], _allowed_colors, shape=(3,))
        door_pos = jax.random.randint(keys[2], shape=(), minval=1, maxval=params.height - 1)

        grid = two_rooms(params.height, params.width)
        grid = grid.at[door_pos, params.width // 2].set(TILES_REGISTRY[Tiles.DOOR_LOCKED, door_color])
        # mask out positions after the wall, so that agent and key are always on the same side
        # WARN: this is a bit expensive, judging by the FPS benchmark
        key_mask = coordinates_mask(grid, (params.height, params.width // 2), comparison_fn=jnp.less)
        key_coords = sample_coordinates(keys[3], grid, num=1, mask=key_mask).squeeze()

        obj_mask = coordinates_mask(grid, (0, params.width // 2 + 1), comparison_fn=jnp.greater_equal)
        obj_coords = sample_coordinates(keys[4], grid, num=1, mask=obj_mask).squeeze()

        grid = grid.at[key_coords[0], key_coords[1]].set(TILES_REGISTRY[Tiles.KEY, door_color])
        grid = grid.at[obj_coords[0], obj_coords[1]].set(TILES_REGISTRY[pickup_obj, pickup_obj_color])
        grid = grid.at[door_pos, params.width // 2 - 1].set(TILES_REGISTRY[block_obj, block_obj_color])

        agent = AgentState(
            position=sample_coordinates(keys[5], grid, num=1, mask=key_mask).squeeze(),
            direction=sample_direction(keys[6]),
        )
        goal_encoding = AgentHoldGoal(tile=TILES_REGISTRY[pickup_obj, pickup_obj_color]).encode()

        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=goal_encoding,
            rule_encoding=_rule_encoding,
            carry=EnvCarry(),
        )
        return state



from __future__ import annotations

from typing import Generic, TypeVar, Union

import jax
import jax.numpy as jnp
from flax import struct
from typing_extensions import TypeAlias

from .core.constants import TILES_REGISTRY, Colors, Tiles


class RuleSet(struct.PyTreeNode):
    goal: jax.Array
    rules: jax.Array
    init_tiles: jax.Array


GridState: TypeAlias = jax.Array
Tile: TypeAlias = jax.Array
IntOrArray: TypeAlias = Union[int, jax.Array]
EnvCarryT = TypeVar("EnvCarryT")


class AgentState(struct.PyTreeNode):
    position: jax.Array = struct.field(default_factory=lambda: jnp.asarray((0, 0)))
    direction: jax.Array = struct.field(default_factory=lambda: jnp.asarray(0))
    pocket: jax.Array = struct.field(default_factory=lambda: TILES_REGISTRY[Tiles.EMPTY, Colors.EMPTY])


class EnvCarry(struct.PyTreeNode): ...


class State(struct.PyTreeNode, Generic[EnvCarryT]):
    key: jax.Array
    step_num: jax.Array

    grid: GridState
    agent: AgentState
    goal_encoding: jax.Array
    rule_encoding: jax.Array

    carry: EnvCarryT


class StepType(jnp.uint8):
    FIRST: jax.Array = jnp.asarray(0, dtype=jnp.uint8)
    MID: jax.Array = jnp.asarray(1, dtype=jnp.uint8)
    LAST: jax.Array = jnp.asarray(2, dtype=jnp.uint8)


class TimeStep(struct.PyTreeNode, Generic[EnvCarryT]):
    state: State[EnvCarryT]

    step_type: StepType
    reward: jax.Array
    discount: jax.Array
    observation: Union[jax.Array, dict[str, jax.Array]]

    def first(self):
        return self.step_type == StepType.FIRST

    def mid(self):
        return self.step_type == StepType.MID

    def last(self):
        return self.step_type == StepType.LAST



from __future__ import annotations

import bz2
import os
import pickle
import urllib.request
from typing import Callable, Union

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from flax import struct
from tqdm.auto import tqdm

from .types import RuleSet

HF_REPO_ID = os.environ.get("XLAND_MINIGRID_HF_REPO_ID", "Howuhh/xland_minigrid")
DATA_PATH = os.environ.get("XLAND_MINIGRID_DATA", os.path.expanduser("~/.xland_minigrid"))

NAME2HFFILENAME = {
    "trivial-21k": "trivial_21k",
    "small-1m": "small_1m_v3",
    "medium-1m": "medium_1m_v3",
    "medium-3m": "medium_3m_v3",
    "high-1m": "high_1m_v3",
    "high-3m": "high_3m_v3",
}


# jit compatible sampling and indexing!
# You can implement your custom curriculums based on this class.
class Benchmark(struct.PyTreeNode):
    goals: jax.Array
    rules: jax.Array
    init_tiles: jax.Array
    num_rules: jax.Array

    def num_rulesets(self) -> int:
        return len(self.goals)

    def get_ruleset(self, ruleset_id: Union[int, jax.Array]) -> RuleSet:
        return get_ruleset(self.goals, self.rules, self.init_tiles, ruleset_id)

    def sample_ruleset(self, key: jax.Array) -> RuleSet:
        ruleset_id = jax.random.randint(key, shape=(), minval=0, maxval=self.num_rulesets())
        return self.get_ruleset(ruleset_id)

    def shuffle(self, key: jax.Array) -> Benchmark:
        idxs = jax.random.permutation(key, jnp.arange(len(self.num_rules)))
        return jtu.tree_map(lambda a: a[idxs], self)

    def split(self, prop: float) -> tuple[Benchmark, Benchmark]:
        idx = round(len(self.num_rules) * prop)
        bench1 = jtu.tree_map(lambda a: a[:idx], self)
        bench2 = jtu.tree_map(lambda a: a[idx:], self)
        return bench1, bench2

    def filter_split(self, fn: Callable[[jax.Array, jax.Array], bool]) -> tuple[Benchmark, Benchmark]:
        # fn(single_goal, single_rules) -> bool
        mask = jax.vmap(fn)(self.goals, self.rules)
        bench1 = jtu.tree_map(lambda a: a[mask], self)
        bench2 = jtu.tree_map(lambda a: a[~mask], self)
        return bench1, bench2


def load_benchmark_from_path(path: str) -> Benchmark:
    benchmark_dict = load_bz2_pickle(path)
    benchmark = Benchmark(
        goals=benchmark_dict["goals"],
        rules=benchmark_dict["rules"],
        init_tiles=benchmark_dict["init_tiles"],
        num_rules=benchmark_dict["num_rules"],
    )
    return benchmark


def load_benchmark(name: str) -> Benchmark:
    if name not in NAME2HFFILENAME:
        raise RuntimeError(f"Unknown benchmark. Registered: {registered_benchmarks()}")

    os.makedirs(DATA_PATH, exist_ok=True)

    path = os.path.join(DATA_PATH, NAME2HFFILENAME[name])
    if not os.path.exists(path):
        _download_from_hf(HF_REPO_ID, NAME2HFFILENAME[name])

    return load_benchmark_from_path(path)


def registered_benchmarks() -> tuple[str, ...]:
    return tuple(NAME2HFFILENAME.keys())


def _download_from_hf(repo_id: str, filename: str) -> None:
    dataset_url = f"https://huggingface.co/datasets/{repo_id}/resolve/main/{filename}"

    save_path = os.path.join(DATA_PATH, filename)
    print(f"Downloading benchmark data: {dataset_url} to {DATA_PATH}")

    with tqdm(unit="B", unit_scale=True, miniters=1, desc="Progress") as t:

        def progress_hook(block_num=1, block_size=1, total_size=None):
            if total_size is not None:
                t.total = total_size
            t.update(block_num * block_size - t.n)

        urllib.request.urlretrieve(dataset_url, save_path, reporthook=progress_hook)

    if not os.path.exists(os.path.join(DATA_PATH, filename)):
        raise IOError(f"Failed to download benchmark data from {dataset_url}")


def get_ruleset(
    goals: jax.Array,
    rules: jax.Array,
    init_tiles: jax.Array,
    ruleset_id: Union[int, jax.Array],
) -> RuleSet:
    goal = jax.lax.dynamic_index_in_dim(goals, ruleset_id, keepdims=False)
    rules = jax.lax.dynamic_index_in_dim(rules, ruleset_id, keepdims=False)
    init_tiles = jax.lax.dynamic_index_in_dim(init_tiles, ruleset_id, keepdims=False)

    return RuleSet(goal=goal, rules=rules, init_tiles=init_tiles)


def save_bz2_pickle(ruleset: dict[str, jax.Array], path: str, protocol: int = -1) -> None:
    with bz2.open(path, "wb") as f:
        pickle.dump(ruleset, f, protocol=protocol)


def load_bz2_pickle(path: str) -> dict[str, jax.Array]:
    with bz2.open(path, "rb") as f:
        ruleset = pickle.load(f)
    return ruleset



from __future__ import annotations

import jax
import jax.numpy as jnp

from ...core.constants import TILES_REGISTRY, Colors, Tiles
from ...core.goals import TileOnPositionGoal
from ...core.grid import coordinates_mask, sample_coordinates, sample_direction, two_rooms
from ...core.rules import EmptyRule
from ...environment import Environment, EnvParams
from ...types import AgentState, EnvCarry, State

# colors like in the original minigrid
_allowed_colors = jnp.array(
    (
        Colors.RED,
        Colors.GREEN,
        Colors.BLUE,
        Colors.PURPLE,
        Colors.YELLOW,
        Colors.GREY,
    )
)
_rule_encoding = EmptyRule().encode()[None, ...]


class Unlock(Environment[EnvParams, EnvCarry]):
    def default_params(self, **kwargs) -> EnvParams:
        params = EnvParams(height=6, width=11)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # formula directly taken from MiniGrid
            params = params.replace(max_steps=8 * params.height**2)
        return params

    def _generate_problem(self, params: EnvParams, key: jax.Array) -> State[EnvCarry]:
        key, *keys = jax.random.split(key, num=5)

        color = jax.random.choice(keys[0], _allowed_colors)
        door_pos = jax.random.randint(keys[1], shape=(), minval=1, maxval=params.height - 1)

        grid = two_rooms(params.height, params.width)
        grid = grid.at[door_pos, params.width // 2].set(TILES_REGISTRY[Tiles.DOOR_LOCKED, color])

        # mask out positions after the wall, so that agent and key are always on the same side
        # WARN: this is a bit expensive, judging by the FPS benchmark
        mask = coordinates_mask(grid, (params.height, params.width // 2), comparison_fn=jnp.less)
        key_coords, agent_coords = sample_coordinates(keys[2], grid, num=2, mask=mask)

        grid = grid.at[key_coords[0], key_coords[1]].set(TILES_REGISTRY[Tiles.KEY, color])

        agent = AgentState(position=agent_coords, direction=sample_direction(keys[3]))
        goal_encoding = TileOnPositionGoal(
            tile=TILES_REGISTRY[Tiles.DOOR_OPEN, color],
            position=jnp.array((door_pos, params.width // 2)),
        ).encode()

        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=goal_encoding,
            rule_encoding=_rule_encoding,
            carry=EnvCarry(),
        )
        return state



from __future__ import annotations

import jax
import jax.numpy as jnp
from flax import struct

from ...core.actions import take_action
from ...core.constants import TILES_REGISTRY, Colors, Tiles
from ...core.goals import EmptyGoal
from ...core.grid import equal, horizontal_line, rectangle, room, vertical_line
from ...core.observation import transparent_field_of_view
from ...core.rules import EmptyRule
from ...environment import Environment, EnvParams
from ...types import AgentState, IntOrArray, State, StepType, TimeStep  # , EnvCarry

_goal_encoding = EmptyGoal().encode()
_rule_encoding = EmptyRule().encode()[None, ...]

_objects = jnp.array(
    (
        (Tiles.BALL, Colors.GREEN),
        (Tiles.KEY, Colors.GREEN),
    ),
    dtype=jnp.uint8,
)

_wall_tile = TILES_REGISTRY[Tiles.WALL, Colors.GREY]
_floor_tile = TILES_REGISTRY[Tiles.FLOOR, Colors.BLACK]


# It can be made to be a goal, but for demonstration
# purposes (how to use carry) we decided to leave it as is
# class MemoryEnvCarry(EnvCarry):
class MemoryEnvCarry(struct.PyTreeNode):
    success_pos: jax.Array
    failure_pos: jax.Array


# TODO: Random corridor length is a bit problematic due to the dynamic slicing.
class Memory(Environment[EnvParams, MemoryEnvCarry]):
    def default_params(self, **kwargs) -> EnvParams:
        params = EnvParams(height=7, width=13, view_size=3)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # formula directly taken from MiniGrid
            params = params.replace(max_steps=5 * params.width**2)
        return params

    def _generate_problem(self, params: EnvParams, key: jax.Array) -> State[MemoryEnvCarry]:
        key, corridor_key, agent_key, mem_key, place_key = jax.random.split(key, num=5)

        corridor_length = params.width - 6
        corridor_end = 4 + corridor_length

        # setting up the world
        grid = room(params.height, params.width)
        grid = rectangle(grid, 0, 1, 5, 5, _wall_tile)
        grid = grid.at[params.height // 2, 4].set(_floor_tile)

        grid = horizontal_line(grid, 4, 2, corridor_length, _wall_tile)
        grid = horizontal_line(grid, 4, 4, corridor_length, _wall_tile)
        grid = grid.at[1, 3 + corridor_length].set(_wall_tile)
        grid = grid.at[5, 3 + corridor_length].set(_wall_tile)
        grid = vertical_line(grid, corridor_end + 1, 0, params.height, _wall_tile)

        # object to memorize
        obj_to_memorize = jax.random.choice(mem_key, _objects, shape=())
        grid = grid.at[2, 1].set(obj_to_memorize)

        # objects to choose
        sides = jax.random.randint(place_key, shape=(), minval=0, maxval=2)
        grid = jax.lax.select(
            sides,
            grid.at[1, corridor_end].set(_objects[0]).at[5, corridor_end].set(_objects[1]),
            grid.at[1, corridor_end].set(_objects[1]).at[5, corridor_end].set(_objects[0]),
        )

        # choosing success and failure positions
        obj_equal_to_upper = equal(obj_to_memorize, grid[1, corridor_end])
        success_pos = jax.lax.select(
            obj_equal_to_upper,
            jnp.array((2, corridor_end)),
            jnp.array((4, corridor_end)),
        )
        failure_pos = jax.lax.select(
            obj_equal_to_upper,
            jnp.array((4, corridor_end)),
            jnp.array((2, corridor_end)),
        )

        # sampling agent position
        agent_x = jax.random.randint(agent_key, shape=(), minval=1, maxval=corridor_end)
        agent = AgentState(
            position=jnp.array((3, agent_x)),
            direction=jnp.asarray(1),
        )
        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=_goal_encoding,
            rule_encoding=_rule_encoding,
            carry=MemoryEnvCarry(success_pos=success_pos, failure_pos=failure_pos),
        )
        return state

    def step(
        self, params: EnvParams, timestep: TimeStep[MemoryEnvCarry], action: IntOrArray
    ) -> TimeStep[MemoryEnvCarry]:
        # disabling pick_up action
        action = jax.lax.select(
            jnp.equal(action, 3),
            jnp.asarray(5),
            action,
        ).astype(jnp.uint8)

        new_grid, new_agent, _ = take_action(timestep.state.grid, timestep.state.agent, action)

        new_state = timestep.state.replace(grid=new_grid, agent=new_agent, step_num=timestep.state.step_num + 1)
        new_observation = transparent_field_of_view(new_state.grid, new_state.agent, params.view_size, params.view_size)

        truncated = new_state.step_num == params.max_steps
        terminated = jnp.logical_or(
            jnp.array_equal(new_agent.position, new_state.carry.success_pos),
            jnp.array_equal(new_agent.position, new_state.carry.failure_pos),
        )
        assert params.max_steps is not None
        reward = jax.lax.select(
            jnp.array_equal(new_agent.position, new_state.carry.success_pos),
            1.0 - 0.9 * (new_state.step_num / params.max_steps),
            0.0,
        )
        step_type = jax.lax.select(terminated | truncated, StepType.LAST, StepType.MID)
        discount = jax.lax.select(terminated, jnp.asarray(0.0), jnp.asarray(1.0))

        timestep = TimeStep(
            state=new_state,
            step_type=step_type,
            reward=reward,
            discount=discount,
            observation=new_observation,
        )
        return timestep



from .minigrid.blockedunlockpickup import BlockedUnlockPickUp
from .minigrid.doorkey import DoorKey
from .minigrid.empty import Empty, EmptyRandom
from .minigrid.fourrooms import FourRooms
from .minigrid.lockedroom import LockedRoom
from .minigrid.memory import Memory
from .minigrid.playground import Playground
from .minigrid.unlock import Unlock
from .minigrid.unlockpickup import UnlockPickUp
from .xland import XLandMiniGrid

__all__ = [
    "BlockedUnlockPickUp",
    "DoorKey",
    "Empty",
    "EmptyRandom",
    "FourRooms",
    "LockedRoom",
    "Memory",
    "Playground",
    "Unlock",
    "UnlockPickUp",
    "XLandMiniGrid",
]



from __future__ import annotations

import jax
import jax.numpy as jnp
from flax import struct

from ..core.constants import TILES_REGISTRY, Colors, Tiles
from ..core.goals import EmptyGoal
from ..core.grid import (
    equal,
    four_rooms,
    horizontal_line,
    nine_rooms,
    room,
    sample_coordinates,
    sample_direction,
    two_rooms,
    vertical_line,
)
from ..core.rules import EmptyRule
from ..environment import Environment, EnvParams
from ..types import AgentState, EnvCarry, GridState, RuleSet, State

_empty_ruleset = RuleSet(
    goal=EmptyGoal().encode(),
    rules=EmptyRule().encode()[None, ...],
    init_tiles=jnp.array(((TILES_REGISTRY[Tiles.EMPTY, Colors.EMPTY],))),
)

_wall_tile = TILES_REGISTRY[Tiles.WALL, Colors.GREY]
# colors for doors between rooms
_allowed_colors = jnp.array(
    (
        Colors.RED,
        Colors.GREEN,
        Colors.BLUE,
        Colors.PURPLE,
        Colors.YELLOW,
        Colors.GREY,
    )
)


# helper functions to generate various maps, inspired by the common minigrid layouts
# TODO: all worlds should be square
def generate_room(key: jax.Array, height: int, width: int) -> tuple[jax.Array, GridState]:
    grid = room(height, width)
    return key, grid


def generate_two_rooms(key: jax.Array, height: int, width: int) -> tuple[jax.Array, GridState]:
    key, color_key, door_key = jax.random.split(key, num=3)

    color = jax.random.choice(color_key, _allowed_colors)
    door_pos = jax.random.randint(door_key, shape=(), minval=1, maxval=height - 1)

    grid = two_rooms(height, width)
    grid = grid.at[door_pos, width // 2].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, color])

    return key, grid


def generate_four_rooms(key: jax.Array, height: int, width: int) -> tuple[jax.Array, GridState]:
    key, doors_key, colors_key = jax.random.split(key, num=3)

    doors_offsets = jax.random.randint(doors_key, shape=(4,), minval=1, maxval=height // 2)
    colors = jax.random.choice(colors_key, _allowed_colors, shape=(4,))

    grid = four_rooms(height, width)
    grid = grid.at[height // 2, doors_offsets[0]].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, colors[0]])
    grid = grid.at[height // 2, width // 2 + doors_offsets[1]].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, colors[1]])
    grid = grid.at[doors_offsets[2], width // 2].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, colors[2]])
    grid = grid.at[height // 2 + doors_offsets[3], width // 2].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, colors[3]])

    return key, grid


def generate_six_rooms(key: jax.Array, height: int, width: int) -> tuple[jax.Array, GridState]:
    key, colors_key = jax.random.split(key)

    grid = room(height, width)
    grid = vertical_line(grid, width // 2 - 2, 0, height, _wall_tile)
    grid = vertical_line(grid, width // 2 + 2, 0, height, _wall_tile)

    for i in range(1, 3):
        grid = horizontal_line(grid, 0, i * (height // 3), width // 2 - 2, _wall_tile)
        grid = horizontal_line(grid, width // 2 + 2, i * (height // 3), width // 2 - 2, _wall_tile)

    doors_idxs = (
        # left doors
        (height // 2 - (height // 3), width // 2 - 2),
        (height // 2, width // 2 - 2),
        (height // 2 + (height // 3), width // 2 - 2),
        # right doors
        (height // 2 - (height // 3), width // 2 + 2),
        (height // 2, width // 2 + 2),
        (height // 2 + (height // 3), width // 2 + 2),
    )
    colors = jax.random.choice(colors_key, _allowed_colors, shape=(6,))

    for i in range(6):
        grid = grid.at[doors_idxs[i][0], doors_idxs[i][1]].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, colors[i]])

    return key, grid


def generate_nine_rooms(key: jax.Array, height: int, width: int) -> tuple[jax.Array, GridState]:
    # valid sizes should follow 3 * x + 4: 7, 10, 13, 16, 19, 22, 25, 28, 31, ...
    # (size - 4) % 3 == 0
    key, doors_key, colors_key = jax.random.split(key, num=3)
    roomW, roomH = width // 3, height // 3

    grid = nine_rooms(height, width)
    # assuming that rooms are square!
    door_coords = jax.random.randint(doors_key, shape=(12,), minval=1, maxval=roomW)
    colors = jax.random.choice(colors_key, _allowed_colors, shape=(12,))

    # adapted from minigrid playground
    door_idx = 0
    for i in range(0, 3):
        for j in range(0, 3):
            xL = i * roomW
            yT = j * roomH
            xR = xL + roomW
            yB = yT + roomH

            if i + 1 < 3:
                grid = grid.at[yT + door_coords[door_idx], xR].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, colors[door_idx]])
                door_idx = door_idx + 1

            if j + 1 < 3:
                grid = grid.at[yB, xL + door_coords[door_idx]].set(TILES_REGISTRY[Tiles.DOOR_CLOSED, colors[door_idx]])
                door_idx = door_idx + 1

    return key, grid


class XLandEnvParams(EnvParams):
    # you can vmap on rulesets for multi-task/meta learning
    ruleset: RuleSet = struct.field(pytree_node=True, default=_empty_ruleset)
    # experimental (can not vmap on it)
    grid_type: int = struct.field(pytree_node=False, default="1R")


class XLandMiniGrid(Environment[XLandEnvParams, EnvCarry]):
    def default_params(self, **kwargs) -> XLandEnvParams:
        params = XLandEnvParams(view_size=5)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # this is just a heuristic to prevent brute force in one episode,
            # so that agent need to remember what he tried in previous episodes.
            # If this is too small, change it or increase number of trials (NB: these are not equivalent).
            params = params.replace(max_steps=3 * (params.height * params.width))
        return params

    def _generate_problem(self, params: XLandEnvParams, key: jax.Array) -> State[EnvCarry]:
        # WARN: we can make this compatible with jit (to vmap on different layouts during training),
        # but it will probably be very costly, as lax.switch will generate all layouts during reset under vmap
        # TODO: experiment with this under jit, does it possible to make it jit-compatible without overhead?
        if params.grid_type == "R1":
            key, grid = generate_room(key, params.height, params.width)
        elif params.grid_type == "R2":
            key, grid = generate_two_rooms(key, params.height, params.width)
        elif params.grid_type == "R4":
            key, grid = generate_four_rooms(key, params.height, params.width)
        elif params.grid_type == "R6":
            key, grid = generate_six_rooms(key, params.height, params.width)
        elif params.grid_type == "R9":
            key, grid = generate_nine_rooms(key, params.height, params.width)
        else:
            # WARN: will not work under jit!
            raise RuntimeError('Unknown grid type, should be one of: ["R1", "R2", "R4", "R6", "R9"]')

        num_objects = len(params.ruleset.init_tiles)
        objects = params.ruleset.init_tiles

        key, coords_key, dir_key = jax.random.split(key, num=3)
        positions = sample_coordinates(coords_key, grid, num=num_objects + 1)
        for i in range(num_objects):
            # we ignore empty tiles, as they are just paddings to the same shape
            grid = jax.lax.select(
                equal(objects[i], TILES_REGISTRY[Tiles.EMPTY, Colors.EMPTY]),
                grid,
                grid.at[positions[i][0], positions[i][1]].set(objects[i]),
            )

        agent = AgentState(position=positions[-1], direction=sample_direction(dir_key))
        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=params.ruleset.goal,
            rule_encoding=params.ruleset.rules,
            carry=EnvCarry(),
        )
        return state



from __future__ import annotations

import jax
import jax.numpy as jnp

from ...core.constants import TILES_REGISTRY, Colors, Tiles
from ...core.goals import AgentOnTileGoal
from ...core.grid import room, sample_coordinates, sample_direction
from ...core.rules import EmptyRule
from ...environment import Environment, EnvParams
from ...types import AgentState, EnvCarry, State

# goals and rules are hardcoded for minigrid envs
_goal_encoding = AgentOnTileGoal(tile=TILES_REGISTRY[Tiles.GOAL, Colors.GREEN]).encode()
_rule_encoding = EmptyRule().encode()[None, ...]


class Empty(Environment[EnvParams, EnvCarry]):
    def default_params(self, **kwargs) -> EnvParams:
        params = EnvParams(height=9, width=9)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # formula directly taken from MiniGrid
            params = params.replace(max_steps=4 * (params.height * params.width))
        return params

    def _generate_problem(self, params: EnvParams, key: jax.Array) -> State[EnvCarry]:
        grid = room(params.height, params.width)

        grid = grid.at[params.height - 2, params.width - 2].set(TILES_REGISTRY[Tiles.GOAL, Colors.GREEN])
        agent = AgentState(position=jnp.array((1, 1)), direction=jnp.asarray(1))

        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=_goal_encoding,
            rule_encoding=_rule_encoding,
            carry=EnvCarry(),
        )
        return state


class EmptyRandom(Environment[EnvParams, EnvCarry]):
    def default_params(self, **kwargs) -> EnvParams:
        params = EnvParams(height=9, width=9)
        params = params.replace(**kwargs)

        if params.max_steps is None:
            # formula directly taken from MiniGrid
            params = params.replace(max_steps=4 * (params.height * params.width))
        return params

    def _generate_problem(self, params: EnvParams, key: jax.Array) -> State[EnvCarry]:
        key, pos_key, dir_key = jax.random.split(key, num=3)

        grid = room(params.height, params.width)
        grid = grid.at[params.height - 2, params.width - 2].set(TILES_REGISTRY[Tiles.GOAL, Colors.GREEN])

        agent = AgentState(
            position=sample_coordinates(pos_key, grid, num=1).squeeze(), direction=sample_direction(dir_key)
        )
        state = State(
            key=key,
            step_num=jnp.asarray(0),
            grid=grid,
            agent=agent,
            goal_encoding=_goal_encoding,
            rule_encoding=_rule_encoding,
            carry=EnvCarry(),
        )
        return state



import jax.numpy as jnp
from flax import struct

NUM_ACTIONS = 6

# GRID: [tile, color]
NUM_LAYERS = 2
NUM_TILES = 13
NUM_COLORS = 12


# enums, kinda...
class Tiles(struct.PyTreeNode):
    EMPTY: int = struct.field(pytree_node=False, default=0)
    FLOOR: int = struct.field(pytree_node=False, default=1)
    WALL: int = struct.field(pytree_node=False, default=2)
    BALL: int = struct.field(pytree_node=False, default=3)
    SQUARE: int = struct.field(pytree_node=False, default=4)
    PYRAMID: int = struct.field(pytree_node=False, default=5)
    GOAL: int = struct.field(pytree_node=False, default=6)
    KEY: int = struct.field(pytree_node=False, default=7)
    DOOR_LOCKED: int = struct.field(pytree_node=False, default=8)
    DOOR_CLOSED: int = struct.field(pytree_node=False, default=9)
    DOOR_OPEN: int = struct.field(pytree_node=False, default=10)
    HEX: int = struct.field(pytree_node=False, default=11)
    STAR: int = struct.field(pytree_node=False, default=12)


class Colors(struct.PyTreeNode):
    EMPTY: int = struct.field(pytree_node=False, default=0)
    RED: int = struct.field(pytree_node=False, default=1)
    GREEN: int = struct.field(pytree_node=False, default=2)
    BLUE: int = struct.field(pytree_node=False, default=3)
    PURPLE: int = struct.field(pytree_node=False, default=4)
    YELLOW: int = struct.field(pytree_node=False, default=5)
    GREY: int = struct.field(pytree_node=False, default=6)
    BLACK: int = struct.field(pytree_node=False, default=7)
    ORANGE: int = struct.field(pytree_node=False, default=8)
    WHITE: int = struct.field(pytree_node=False, default=9)
    BROWN: int = struct.field(pytree_node=False, default=10)
    PINK: int = struct.field(pytree_node=False, default=11)


# Only ~100 combinations so far, better to preallocate them
TILES_REGISTRY = []
for tile_id in range(NUM_TILES):
    for color_id in range(NUM_COLORS):
        TILES_REGISTRY.append((tile_id, color_id))
TILES_REGISTRY = jnp.array(TILES_REGISTRY, dtype=jnp.uint8).reshape(NUM_TILES, NUM_COLORS, -1)


DIRECTIONS = jnp.array(
    (
        (-1, 0),  # Up
        (0, 1),  # Right
        (1, 0),  # Down
        (0, -1),  # Left
    )
)


WALKABLE = jnp.array(
    (
        Tiles.FLOOR,
        Tiles.GOAL,
        Tiles.DOOR_OPEN,
    )
)

PICKABLE = jnp.array(
    (
        Tiles.BALL,
        Tiles.SQUARE,
        Tiles.PYRAMID,
        Tiles.KEY,
        Tiles.HEX,
        Tiles.STAR,
    )
)

FREE_TO_PUT_DOWN = jnp.array((Tiles.FLOOR,))

LOS_BLOCKING = jnp.array(
    (
        Tiles.WALL,
        Tiles.DOOR_LOCKED,
        Tiles.DOOR_CLOSED,
    )
)



from __future__ import annotations

import abc
from typing import Any, Generic, Optional, TypeVar, Union

import jax
import jax.numpy as jnp
import numpy as np
from flax import struct

from .core.actions import take_action
from .core.constants import NUM_ACTIONS, NUM_LAYERS
from .core.goals import check_goal
from .core.observation import transparent_field_of_view
from .core.rules import check_rule
from .rendering.rgb_render import render as rgb_render
from .rendering.text_render import render as text_render
from .types import EnvCarryT, IntOrArray, State, StepType, TimeStep


class EnvParams(struct.PyTreeNode):
    # WARN: pytree_node=False, so you CAN NOT vmap on them!
    # You can add pytree node params, but be careful and
    # test that your code will work under jit.
    # Spoiler: probably it will not :(
    height: int = struct.field(pytree_node=False, default=9)
    width: int = struct.field(pytree_node=False, default=9)
    view_size: int = struct.field(pytree_node=False, default=7)
    max_steps: Optional[int] = struct.field(pytree_node=False, default=None)
    render_mode: str = struct.field(pytree_node=False, default="rgb_array")


EnvParamsT = TypeVar("EnvParamsT", bound="EnvParams")


class Environment(abc.ABC, Generic[EnvParamsT, EnvCarryT]):
    @abc.abstractmethod
    def default_params(self, **kwargs: Any) -> EnvParamsT: ...

    def num_actions(self, params: EnvParamsT) -> int:
        return int(NUM_ACTIONS)

    def observation_shape(self, params: EnvParamsT) -> Union[tuple[int, int, int], dict[str, Any]]:
        return params.view_size, params.view_size, NUM_LAYERS

    @abc.abstractmethod
    def _generate_problem(self, params: EnvParamsT, key: jax.Array) -> State[EnvCarryT]: ...

    def reset(self, params: EnvParamsT, key: jax.Array) -> TimeStep[EnvCarryT]:
        state = self._generate_problem(params, key)
        timestep = TimeStep(
            state=state,
            step_type=StepType.FIRST,
            reward=jnp.asarray(0.0),
            discount=jnp.asarray(1.0),
            observation=transparent_field_of_view(state.grid, state.agent, params.view_size, params.view_size),
        )
        return timestep

    # Why timestep + state at once, and not like in Jumanji? To be able to do autoresets in gym and envpools styles
    def step(self, params: EnvParamsT, timestep: TimeStep[EnvCarryT], action: IntOrArray) -> TimeStep[EnvCarryT]:
        new_grid, new_agent, changed_position = take_action(timestep.state.grid, timestep.state.agent, action)
        new_grid, new_agent = check_rule(timestep.state.rule_encoding, new_grid, new_agent, action, changed_position)

        new_state = timestep.state.replace(
            grid=new_grid,
            agent=new_agent,
            step_num=timestep.state.step_num + 1,
        )
        new_observation = transparent_field_of_view(new_state.grid, new_state.agent, params.view_size, params.view_size)

        # checking for termination or truncation, choosing step type
        terminated = check_goal(new_state.goal_encoding, new_state.grid, new_state.agent, action, changed_position)

        assert params.max_steps is not None
        truncated = jnp.equal(new_state.step_num, params.max_steps)

        reward = jax.lax.select(terminated, 1.0 - 0.9 * (new_state.step_num / params.max_steps), 0.0)

        step_type = jax.lax.select(terminated | truncated, StepType.LAST, StepType.MID)
        discount = jax.lax.select(terminated, jnp.asarray(0.0), jnp.asarray(1.0))

        timestep = TimeStep(
            state=new_state,
            step_type=step_type,
            reward=reward,
            discount=discount,
            observation=new_observation,
        )
        return timestep

    def render(self, params: EnvParamsT, timestep: TimeStep[EnvCarryT]) -> Union[np.ndarray, str]:
        if params.render_mode == "rgb_array":
            return rgb_render(np.asarray(timestep.state.grid), timestep.state.agent, params.view_size)
        elif params.render_mode == "rich_text":
            return text_render(timestep.state.grid, timestep.state.agent)
        else:
            raise RuntimeError("Unknown render mode. Should be one of: ['rgb_array', 'rich_text']")



from __future__ import annotations

import abc
from typing import Union

import jax
import jax.numpy as jnp
from flax import struct

from ..types import AgentState, GridState
from .constants import TILES_REGISTRY, Colors, Tiles
from .grid import equal, get_neighbouring_tiles, pad_along_axis

MAX_RULE_ENCODING_LEN = 6 + 1  # +1 for idx
NUM_RULES = 12


# this is very costly, will evaluate all rules under vmap. Submit a PR if you know how to do it better!
# In general, we need a way to select specific function/class based on ID number.
# We can not just decode without evaluation, as then return type will be different between branches
def check_rule(
    encodings: jax.Array, grid: GridState, agent: AgentState, action: Union[int, jax.Array], position: jax.Array
) -> tuple[GridState, AgentState]:
    def _check(carry, encoding):
        grid, agent = carry
        grid, agent = jax.lax.switch(
            encoding[0],
            (
                # empty rule first, we use them as paddings during benchmark generation
                lambda: EmptyRule.decode(encoding)(grid, agent, action, position),
                lambda: AgentHoldRule.decode(encoding)(grid, agent, action, position),
                lambda: AgentNearRule.decode(encoding)(grid, agent, action, position),
                lambda: TileNearRule.decode(encoding)(grid, agent, action, position),
                # rules for the extended benchmarks
                lambda: TileNearUpRule.decode(encoding)(grid, agent, action, position),
                lambda: TileNearRightRule.decode(encoding)(grid, agent, action, position),
                lambda: TileNearDownRule.decode(encoding)(grid, agent, action, position),
                lambda: TileNearLeftRule.decode(encoding)(grid, agent, action, position),
                lambda: AgentNearUpRule.decode(encoding)(grid, agent, action, position),
                lambda: AgentNearRightRule.decode(encoding)(grid, agent, action, position),
                lambda: AgentNearDownRule.decode(encoding)(grid, agent, action, position),
                lambda: AgentNearLeftRule.decode(encoding)(grid, agent, action, position),
            ),
        )
        return (grid, agent), None

    (grid, agent), _ = jax.lax.scan(_check, (grid, agent), encodings)

    return grid, agent


class BaseRule(struct.PyTreeNode):
    @abc.abstractmethod
    def __call__(
        self, grid: GridState, agent: AgentState, action: Union[int, jax.Array], position: jax.Array
    ) -> tuple[GridState, AgentState]: ...

    @classmethod
    @abc.abstractmethod
    def decode(cls, encoding: jax.Array) -> BaseRule: ...

    @abc.abstractmethod
    def encode(self) -> jax.Array: ...


# TODO: checks on only some actions can produce bugs, but it is faster. If it's going to cause trouble, reconsider.
class EmptyRule(BaseRule):
    def __call__(self, grid, agent, action, position):
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls()

    def encode(self):
        return jnp.zeros(MAX_RULE_ENCODING_LEN, dtype=jnp.uint8)


class AgentHoldRule(BaseRule):
    tile: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _rule_fn(agent):
            agent = jax.lax.cond(
                equal(agent.pocket, self.tile),
                lambda: agent.replace(pocket=self.prod_tile),
                lambda: agent,
            )
            return agent

        # TODO: seems like this check is slowing down, 25k -> 19k FPS
        # agent = _rule_fn(agent)
        agent = jax.lax.cond(
            jnp.equal(action, 3),
            lambda: _rule_fn(agent),
            lambda: agent,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3], prod_tile=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(1), self.tile, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


class AgentNearRule(BaseRule):
    tile: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _rule_fn(grid):
            y, x = agent.position
            up, right, down, left = get_neighbouring_tiles(grid, y, x)

            grid = jax.lax.select(equal(up, self.tile), grid.at[y - 1, x].set(self.prod_tile), grid)
            grid = jax.lax.select(equal(right, self.tile), grid.at[y, x + 1].set(self.prod_tile), grid)
            grid = jax.lax.select(equal(down, self.tile), grid.at[y + 1, x].set(self.prod_tile), grid)
            grid = jax.lax.select(equal(left, self.tile), grid.at[y, x - 1].set(self.prod_tile), grid)
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 0),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3], prod_tile=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(2), self.tile, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


class TileNearRule(BaseRule):
    tile_a: jax.Array
    tile_b: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        tile = grid[position[0], position[1]]
        tile_a = self.tile_a
        tile_b = self.tile_b

        def _rule_fn(grid):
            floor_tile = TILES_REGISTRY[Tiles.FLOOR, Colors.BLACK]
            y, x = position
            up, right, down, left = get_neighbouring_tiles(grid, y, x)

            grid = jax.lax.select(
                (equal(tile, tile_a) & equal(up, tile_b)) | (equal(tile, tile_b) & equal(up, tile_a)),
                grid.at[y - 1, x].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            grid = jax.lax.select(
                (equal(tile, tile_a) & equal(right, tile_b)) | (equal(tile, tile_b) & equal(right, tile_a)),
                grid.at[y, x + 1].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            grid = jax.lax.select(
                (equal(tile, tile_a) & equal(down, tile_b)) | (equal(tile, tile_b) & equal(down, tile_a)),
                grid.at[y + 1, x].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            grid = jax.lax.select(
                (equal(tile, tile_a) & equal(left, tile_b)) | (equal(tile, tile_b) & equal(left, tile_a)),
                grid.at[y, x - 1].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 4) & (equal(tile, self.tile_a) | equal(tile, self.tile_b)),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5], prod_tile=encoding[5:7])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(3), self.tile_a, self.tile_b, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


# tile_b should be one tile up near the tile_a
class TileNearUpRule(BaseRule):
    tile_a: jax.Array
    tile_b: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        tile = grid[position[0], position[1]]

        def _rule_fn(grid):
            floor_tile = TILES_REGISTRY[Tiles.FLOOR, Colors.BLACK]
            y, x = position
            up, _, down, _ = get_neighbouring_tiles(grid, y, x)

            grid = jax.lax.select(
                equal(tile, self.tile_b) & equal(down, self.tile_a),
                grid.at[y + 1, x].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            grid = jax.lax.select(
                equal(tile, self.tile_a) & equal(up, self.tile_b),
                grid.at[y - 1, x].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 4) & (equal(tile, self.tile_a) | equal(tile, self.tile_b)),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5], prod_tile=encoding[5:7])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(4), self.tile_a, self.tile_b, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


class TileNearRightRule(BaseRule):
    tile_a: jax.Array
    tile_b: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        tile = grid[position[0], position[1]]

        def _rule_fn(grid):
            floor_tile = TILES_REGISTRY[Tiles.FLOOR, Colors.BLACK]

            y, x = position
            _, right, _, left = get_neighbouring_tiles(grid, y, x)

            grid = jax.lax.select(
                equal(tile, self.tile_b) & equal(left, self.tile_a),
                grid.at[y, x - 1].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            grid = jax.lax.select(
                equal(tile, self.tile_a) & equal(right, self.tile_b),
                grid.at[y, x + 1].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 4) & (equal(tile, self.tile_a) | equal(tile, self.tile_b)),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5], prod_tile=encoding[5:7])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(5), self.tile_a, self.tile_b, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


class TileNearDownRule(BaseRule):
    tile_a: jax.Array
    tile_b: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        tile = grid[position[0], position[1]]

        def _rule_fn(grid):
            floor_tile = TILES_REGISTRY[Tiles.FLOOR, Colors.BLACK]
            y, x = position
            up, _, down, _ = get_neighbouring_tiles(grid, y, x)

            grid = jax.lax.select(
                equal(tile, self.tile_b) & equal(up, self.tile_a),
                grid.at[y - 1, x].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            grid = jax.lax.select(
                equal(tile, self.tile_a) & equal(down, self.tile_b),
                grid.at[y + 1, x].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 4) & (equal(tile, self.tile_a) | equal(tile, self.tile_b)),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5], prod_tile=encoding[5:7])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(6), self.tile_a, self.tile_b, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


class TileNearLeftRule(BaseRule):
    tile_a: jax.Array
    tile_b: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        tile = grid[position[0], position[1]]

        def _rule_fn(grid):
            floor_tile = TILES_REGISTRY[Tiles.FLOOR, Colors.BLACK]
            y, x = position
            _, right, _, left = get_neighbouring_tiles(grid, y, x)

            grid = jax.lax.select(
                equal(tile, self.tile_b) & equal(right, self.tile_a),
                grid.at[y, x + 1].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            grid = jax.lax.select(
                equal(tile, self.tile_a) & equal(left, self.tile_b),
                grid.at[y, x - 1].set(self.prod_tile).at[y, x].set(floor_tile),
                grid,
            )
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 4) & (equal(tile, self.tile_a) | equal(tile, self.tile_b)),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile_a=encoding[1:3], tile_b=encoding[3:5], prod_tile=encoding[5:7])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(7), self.tile_a, self.tile_b, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


class AgentNearUpRule(BaseRule):
    tile: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _rule_fn(grid):
            y, x = agent.position
            up, _, _, _ = get_neighbouring_tiles(grid, y, x)
            grid = jax.lax.select(
                equal(up, self.tile),
                grid.at[y - 1, x].set(self.prod_tile),
                grid,
            )
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 0) | jnp.equal(action, 4),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3], prod_tile=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(8), self.tile, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


class AgentNearRightRule(BaseRule):
    tile: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _rule_fn(grid):
            y, x = agent.position
            _, right, _, _ = get_neighbouring_tiles(grid, y, x)
            grid = jax.lax.select(
                equal(right, self.tile),
                grid.at[y, x + 1].set(self.prod_tile),
                grid,
            )
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 0) | jnp.equal(action, 4),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3], prod_tile=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(9), self.tile, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


class AgentNearDownRule(BaseRule):
    tile: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _rule_fn(grid):
            y, x = agent.position
            _, _, down, _ = get_neighbouring_tiles(grid, y, x)
            grid = jax.lax.select(
                equal(down, self.tile),
                grid.at[y + 1, x].set(self.prod_tile),
                grid,
            )
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 0) | jnp.equal(action, 4),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3], prod_tile=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(10), self.tile, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)


class AgentNearLeftRule(BaseRule):
    tile: jax.Array
    prod_tile: jax.Array

    def __call__(self, grid, agent, action, position):
        def _rule_fn(grid):
            y, x = agent.position
            _, _, _, left = get_neighbouring_tiles(grid, y, x)
            grid = jax.lax.select(
                equal(left, self.tile),
                grid.at[y, x - 1].set(self.prod_tile),
                grid,
            )
            return grid

        grid = jax.lax.cond(
            jnp.equal(action, 0) | jnp.equal(action, 4),
            lambda: _rule_fn(grid),
            lambda: grid,
        )
        return grid, agent

    @classmethod
    def decode(cls, encoding):
        return cls(tile=encoding[1:3], prod_tile=encoding[3:5])

    def encode(self):
        encoding = jnp.hstack([jnp.asarray(11), self.tile, self.prod_tile], dtype=jnp.uint8)
        return pad_along_axis(encoding, MAX_RULE_ENCODING_LEN)
